-- scan_gen.gs
-- functions for scanner generator


-- auxiliary functions

union [] l = l
union (x:l) l'
	| elem x l' = (union l l')
	| otherwise = x: union l l'

normalize []    l' = []
normalize (a:l) l' | elem a l' = a : normalize l l'
                   | otherwise   = normalize l l'

fixed_point2 :: Eq [a] => ([a] -> [a]) -> [a] -> [a] -> [a]
fixed_point2 f to_process processed
	| y == []  = processed
	| otherwise = fixed_point2 f y (y ++ processed)
			where
			  y = f to_process \\ processed

show_char '\a' = "\\a"
show_char '\t' = "\\t"
show_char '\n' = "\\n"
show_char '\b' = "\\b"
show_char '\v' = "\\v"
show_char '\f' = "\\f"
show_char '\'' = "\\\'"
show_char '\"' = "\\\""
show_char '\\' = "\\\\"
show_char c
	| c<' ' || c>'~' = '\\':show (ord c)
	| otherwise	 = [c]

show_char_q c = '\'':(show_char c ++ "\'")

-- first_syms r is the set of charaters that can begin a string
-- that r can derive.
-- The set is given as a sorted list of disjoint intervals

first_syms (Symbs c1 c2) = [(c1,c2)]
first_syms (Seq r1 r2)
	| nullable r1    = union_syms (first_syms r1) (first_syms r2)
	| otherwise      = first_syms r1
first_syms (Alt r1 r2)   = union_syms (first_syms r1) (first_syms r2)
first_syms (Repeat r)    = first_syms r
first_syms Empty	 = []
first_syms Void		 = []
first_syms (Label _)	 = []

-- merges two sorted lists of disjoint intervals
-- splitting intervals where neccessary

union_syms [] l = l
union_syms l [] = l
union_syms l1@((c11,c12):l1') l2@((c21,c22):l2')
	| c12<c21   = (c11,c12): union_syms l1' l2
	| c22<c11   = (c21,c22): union_syms l1 l2'
	| c11<c21   = (c11,chr (ord c21 - 1))
			: union_syms ((c21,c12):l1') l2
	| c11>c21   = (c21,chr (ord c11 - 1))
			: union_syms l1 ((c11,c22):l2')
	| c12==c22  = (c11,c12): union_syms l1' l2'
	| c12<c22   = (c11,c12)
			: union_syms l1' ((chr (ord c12 + 1),c22):l2')
	| c12>c22   = (c11,c22)
			: union_syms ((chr (ord c22 + 1),c12):l1') l2'

-- next c r is a regular expression that accepts a string s
-- whenever r accepts the string c:s.
-- r must accept strings beginning with c

next _ (Symbs _ _) = Empty
next c (Seq r1 r2)
	| nullable r1 && included c (first_syms r1)
		      && included c (first_syms r2)
		= mAlt (mSeq (next c r1) r2) (next c r2)
	| nullable r1 && included c (first_syms r1)
		= mSeq (next c r1) r2
	| nullable r1 -- by invariant: && included c (first_syms r2)
		= next c r2
	| otherwise
		= mSeq (next c r1) r2
next c (Alt r1 r2)
	| included c (first_syms r1) && included c (first_syms r2)
		= mAlt (next c r1) (next c r2)
	| included c (first_syms r1)
		= next c r1
	| otherwise -- by invariant: included c (first_syms r2)
		= next c r2
next c r@(Repeat r')
	= mSeq (next c r') r

-- included c css tests if the interval cs is included in one of the
-- intervals in css

included _ [] = False
included c ((c1',c2'):css)
	| c>=c1' && c<=c2' = True
	| otherwise	   = included c css

-- labels r finds the set of token labels that occur before any
-- character in r

labels (Symbs _ _) = []
labels (Seq r1 r2)
	| nullable r1 = union (labels r1) (labels r2)
	| otherwise   = labels r1
labels (Alt r1 r2) = union (labels r1) (labels r2)
labels (Repeat r)  = labels r
labels Empty	   = []
labels Void	   = []
labels (Label s)   = [s]

make_regexp (pairs,comments)
	= foldr (\(x,r)-> mAlt (Seq r (Label x))) comments pairs

-- scanner generator functions

scan_gen reg_file scan_name scan_file
  = readFile reg_file
	(\ioerror -> appendChan stdout
				("Can't open " ++ reg_file)
				abort done)
	(\reg_text ->
    let { scanner = scan_gen' reg_text scan_name}
    in
    (writeFile scan_file scanner
	(\ioerror -> appendChan stdout ("Can't write to " ++ scan_file)
		 			abort done)
	done))

scan_gen' reg_text scan_name
  = label_values ++ "\n"
    ++ scan_name ++ " s\n  = scan0 s \"\" \"\" 1\n  where\n"
    ++ scan0_text
    ++ unlines (map make_state reg_env)
  where
    reg = parse_tokens reg_text
    r0 = make_regexp reg
    token_labels = map fst (fst reg)
    first_priority [] = -1
    first_priority ["Comment"] = 0
    first_priority ts = lookup (head (normalize token_labels ts)) label_env 
    label_env = zip token_labels [1..]
    label_values
	= unlines (map (\(x,n)-> x ++ " = " ++ show n) label_env)
    scan0_text
      =    "    scan0 \"\" s0 s1 line = []\n"
	++ "    scan0 s@(c:s\') s0 s1 line\n     =\n"
	++ "     let {\n"
	++ "      reduce = if isSpace c then scan0 s\' s0 (c:s1) (if c==\'\\n\' then line+1 else line)\n"
	++ "               else error (\"Lexical error in line \" ++ show "
			++ "line ++ \" at\\n\"++ head (lines s))}\n"
	++ "     in\n"
	++ "      "
	++ make_bin_search r0 (first_syms r0) '\0' '\255'
	++ "\n\n"

    regs = fixed_point2 next_all [r0] [r0]
    reg_env = zip (reverse regs) [0..] 
    next_all rs = foldr union rs [nub [next c r | (c,c') <- (first_syms r)]
				  | r <- rs]
    make_state (r,n)
	= make_state' (r,n) (first_syms r)

    make_state' (r,n) []
	=    ""
    make_state' (r,n) fs
	=    "    match_regexp_" ++ show n ++ " \"\" s0 s1 line\n"
	  ++ "      = " ++ make_reduce_end (first_priority (labels r)) "line"
		++ "\n"
	  ++ "    match_regexp_" ++ show n ++ " s@(c:s\') s0 s1 line\n     = "
	  ++ "let {\n"
	  ++ "      reduce = " ++ make_reduce (first_priority (labels r)) "line"
	  ++ "}\n     in\n"
	  ++ "      "
	  ++ make_bin_search r fs '\0' '\255'
	  ++ "\n"

    make_reduce (-1) line
	= "error (\"Lexical error in line \" ++ show " ++ line
			++ " ++ \" at\\n\"++ head (lines s))"
    make_reduce 0 line
	= "scan0 s \"\" (s0++s1) "++ line
    make_reduce t line
	= "(" ++ show t ++ ",reverse s0,("++line++",reverse s1 ++ reverse s0))"
	  ++ ": scan0 s \"\" \"\" "++ line

    make_reduce' (-1) line
	= "error (\"Lexical error in line \" ++ show " ++ line
			++ " ++ \" at\\n\"++ head (lines s\'))"
    make_reduce' 0 line
	= "scan0 s\' \"\" (c:s0++s1) "++ line
    make_reduce' t line
	= "(" ++ show t ++ ",reverse (c:s0),("++line++",reverse s1 ++ reverse (c:s0)))"
	  ++ ": scan0 s\' \"\" \"\" "++line

    make_reduce_end (-1) line
	= "error (\"Lexical error in line \" ++ show "++ line
			++ " ++ \" at END\")"
    make_reduce_end 0 line
	= "[]"
    make_reduce_end t line
	= "[(" ++ show t ++ ",reverse s0,("++line++",reverse s1 ++ reverse s0))]"

    make_bin_search r [(c1,c2)] c1' c2'
	| c1==c1' && c2==c2'
		= make_match c1 c2 r
	| c1==c1'
		= "if c<=" ++ show_char_q c2 ++
		  "\n       then " ++ make_match c1 c2 r ++
		  "\n       else reduce"
	| c2==c2'
		= "if c>=" ++ show_char_q c1 ++
		  "\n       then " ++ make_match c1 c2 r ++
		  "\n       else reduce"
	| c1==c2
		= "if c==" ++ show_char_q c1 ++
		  "\n       then " ++ make_match c1 c2 r ++
		  "\n       else reduce"
	| otherwise
		= "if c>=" ++ show_char_q c1 ++
			" && c<=" ++ show_char_q c2 ++
		  "\n       then " ++ make_match c1 c2 r++
		  "\n       else reduce"
    make_bin_search r cs c1' c2'
	= let (cs1,cs2@((c1,_):_)) = splitAt (length cs /2) cs
	  in "if c>=" ++ show_char_q c1 ++
	     "\n       then " ++ make_bin_search r cs2 c1 c2' ++
	     "\n       else " ++ make_bin_search r cs1 c1'
				(chr (ord c1 -1))

    make_match c1 c2 r
	| c1=='\n' &&  c2=='\n'
		= make_match' c1 r "(line+1)"
	| c1<='\n' &&  c2>='\n'
		= make_match' c1 r
		  "(if c==\'\\n\' then line+1 else line)"
	| otherwise
		= make_match' c1 r "line"

    make_match' c r line
	= let nr = next c r in
	    if null (first_syms nr)
	    then make_reduce' (first_priority (labels nr)) line
	    else "match_regexp_" ++ show (lookup nr reg_env)
			++ " s\' (c:s0) s1 "++ line
