-- scanner for token declarations
-- handwritten version

whereT = 'w'
semicolon = ';'
label = 'l'
equals = '='
bar = '|'
char = 'c'
lbracket = '['
rbracket = ']'
string = 's'
dash = '-'
empty = '@'
star = '*'
lpar = '('
rpar = ')'
comment = 'C'

scan = scan_s 1

scan_s line "" = []
scan_s line ('-':'-':s)
	= case (scan_s line s1) of
		[] -> []
		((t,v,(line1,s2)):ts) -> (t,v,(line1,"--"++s0++s2)):ts
	where (s0,s1) = span (/= '\n') s
scan_s line ('\'':s)
	= ('c',s0,(line,'\'':(s0'++"\'"))) : scan_s line s1
	where [(s0,s0'),("","\'"),("",s1)]  = scan_string s
scan_s line ('\"':s)
	= ('s',s0,(line,'\"':s0')) : scan_s line s1
	where ss  = scan_string s
	      s0  = concat (map fst ss)
	      s0' = concat (map snd (init ss))
	      s1  = snd (last ss)
scan_s line ('=':s) = ('=',"=",(line,"=")) : scan_s line s
scan_s line (';':s) = (';',";",(line,";")) : scan_s line s
scan_s line ('|':s) = ('|',"|",(line,"|")) : scan_s line s
scan_s line ('[':s) = ('[',"[",(line,"[")) : scan_s line s
scan_s line (']':s) = (']',"]",(line,"]")) : scan_s line s
scan_s line ('-':s) = ('-',"-",(line,"-")) : scan_s line s
scan_s line ('@':s) = ('@',"@",(line,"@")) : scan_s line s
scan_s line ('*':s) = ('*',"*",(line,"*")) : scan_s line s
scan_s line ('(':s) = ('(',"(",(line,"(")) : scan_s line s
scan_s line (')':s) = (')',")",(line,")")) : scan_s line s
scan_s line ('w':'h':'e':'r':'e':s@(c:_))
	| not (isAlphanum c || c == '_' || c == '\'')
		= ('w',"",(line,"where")) : scan_s line s
scan_s line ('C':'o':'m':'m':'e':'n':'t':s@(c:_))
	| not (isAlphanum c || c == '_' || c == '\'')
		= ('C',"",(line,"Comment")) : scan_s line s
scan_s line s@(c:_)
	| isLower c
		= ('l',s0,(line,s0)) : scan_s line s1
		where
		  (s0, s1) = span (\c -> isAlphanum c || c == '_' || c == '\'') s
scan_s line ('\n':s)
		= case (scan_s (line+1) s) of
			[] -> []
			((t,v,(line1,s0)):ts) -> (t,v,(line1,'\n':s0)):ts
scan_s line (c:s)
	| isSpace c
		= case (scan_s line s) of
			[] -> []
			((t,v,(line1,s0)):ts) -> (t,v,(line1,c:s0)):ts
scan_s line s
	= error ("Lexical error in line " ++ show line ++ " at\n"
		 ++ take 100 s)

scan_string ('\'':s) = [("","\'"),("",s)]
scan_string ('\"':s) = [("","\""),("",s)]
scan_string ('\\':'\\':s)
	= ("\\","\\\\") : scan_string s
scan_string ('\\':'\'':s)
	= ("\'","\\\'") : scan_string s
scan_string ('\\':'\"':s)
	= ("\"","\\\"") : scan_string s
scan_string ('\\':'n':s)
	= ("\n","\\n") : scan_string s
scan_string ('\\':'b':s)
	= ("\b","\\b") : scan_string s
scan_string ('\\':'a':s)
	= ("\a","\\a") : scan_string s
scan_string ('\\':'f':s)
	= ("\f","\\f") : scan_string s
scan_string ('\\':'t':s)
	= ("\t","\\t") : scan_string s
scan_string ('\\':'r':s)
	= ("\r","\\r") : scan_string s
scan_string ('\\':'v':s)
	= ("\v","\\v") : scan_string s
scan_string ('\\':c:s)
	| isDigit c
		= ([chr n],'\\':sn) : scan_string s1
		where (sn,s1) = span isDigit (c:s)
		      n = foldl (\x d -> 10*x+ord d-ord '0') 0 sn
scan_string (c:s)
	= ([c],[c]) : scan_string s
