/* --------------------------------------------------------------------------
 * prims.c:	Copyright (c) Mark P Jones 1991, 1992.   All rights reserved.
 *		See goferite.h for details and conditions of use etc...
 *		Gofer version 2.28 November 1992
 *
 *		Last updated March 93 kh
 *		Added Macintosh support, hyperbolic trig functions,
 *		and hexadecimal operations.
 *
 * Primitive functions, input output etc...
 * if PRIMITIVES_CODE == 0 then the code for PRIMITIVES is excluded: only
 * the primitives table and consChar() parts are retained.
 * ------------------------------------------------------------------------*/

#if PRIMITIVES_CODE


#if MPW
/* Cell's already defined for lists */
#define Cell __Cell__
#include "mac_hdrs.h"
#include "mac_ctype.h"
#undef Cell
#else
#include <ctype.h>
#endif

#if (TURBOC | BCC)
#include <io.h>
#endif
#endif

/* --------------------------------------------------------------------------
 * Local function prototypes:
 * ------------------------------------------------------------------------*/

#if PRIMITIVES_CODE
#define PROTO_PRIM(name)	static Void name Args((StackPtr))
#define primFun(name)		static Void name(root) StackPtr root;
#define primArg(n)		stack(root+n)

/* IMPORTANT: the second element of an update must be written first.
 * this is to deal with the case where an INDIRECT tag is written into
 * a Cell before the second value has been set.  If a garbage collection
 * occurs before the second element was set then the INDIRECTion will be
 * (wrongly) elided and result in chaos.  I know.  It happened to me.
 */

#define update(l,r)		((snd(stack(root))=r),(fst(stack(root))=l))
#define updateRoot(c)		update(INDIRECT,c)
#define updapRoot(l,r)		update(l,r)
#define cantReduce()		evalFails(root)

PROTO_PRIM(primFatbar);
PROTO_PRIM(primFail);
PROTO_PRIM(primSel);
PROTO_PRIM(primIf);
PROTO_PRIM(primStrict);

PROTO_PRIM(primPlusInt);
PROTO_PRIM(primMinusInt);
PROTO_PRIM(primMulInt);
PROTO_PRIM(primDivInt);
PROTO_PRIM(primQuotInt);
PROTO_PRIM(primModInt);
PROTO_PRIM(primRemInt);
PROTO_PRIM(primNegInt);

PROTO_PRIM(primCharToInt);
PROTO_PRIM(primIntToChar);
PROTO_PRIM(primIntToFloat);

PROTO_PRIM(primPlusFloat);
PROTO_PRIM(primMinusFloat);
PROTO_PRIM(primMulFloat);
PROTO_PRIM(primDivFloat);
PROTO_PRIM(primNegFloat);

#if HAS_FLOATS
PROTO_PRIM(primSinFloat);
PROTO_PRIM(primCosFloat);
PROTO_PRIM(primTanFloat);
PROTO_PRIM(primAsinFloat);
PROTO_PRIM(primAcosFloat);
PROTO_PRIM(primAtanFloat);
PROTO_PRIM(primAtan2Float);
PROTO_PRIM(primExpFloat);
PROTO_PRIM(primLogFloat);
PROTO_PRIM(primLog10Float);
PROTO_PRIM(primSqrtFloat);
PROTO_PRIM(primFloatToInt);
#endif

#if HAS_HYPERBOLICS
PROTO_PRIM(primSinhFloat);
PROTO_PRIM(primCoshFloat);
PROTO_PRIM(primTanhFloat);
PROTO_PRIM(primAsinhFloat);
PROTO_PRIM(primAcoshFloat);
PROTO_PRIM(primAtanhFloat);
#endif

PROTO_PRIM(primEqInt);
PROTO_PRIM(primLeInt);

PROTO_PRIM(primEqChar);
PROTO_PRIM(primLeChar);

PROTO_PRIM(primEqFloat);
PROTO_PRIM(primLeFloat);

PROTO_PRIM(primCmp);
PROTO_PRIM(primGenericEq);
PROTO_PRIM(primGenericLe);
PROTO_PRIM(primGenericLt);
PROTO_PRIM(primGenericGe);
PROTO_PRIM(primGenericGt);
PROTO_PRIM(primGenericNe);

PROTO_PRIM(primPrint);
PROTO_PRIM(primNPrint);


/* C Monad Primitives for the Mac -- KH */
#if MAC
PROTO_PRIM(primUnitIO);
PROTO_PRIM(primCUnitIO);
PROTO_PRIM(primBindIO);
PROTO_PRIM(primCBindIO);
PROTO_PRIM(primCCBindIO);

PROTO_PRIM(primTrap);
PROTO_PRIM(primTrapReg);
PROTO_PRIM(primMalloc);
PROTO_PRIM(primAssign);
PROTO_PRIM(primAssignS);
PROTO_PRIM(primAssignC);
PROTO_PRIM(primAssignBlock);
PROTO_PRIM(primDeref);
PROTO_PRIM(primFree);
PROTO_PRIM(primSeq);
PROTO_PRIM(primTrace);

PROTO_PRIM(primButton);
PROTO_PRIM(primGetMouse);
PROTO_PRIM(primLineTo);
PROTO_PRIM(primMoveTo);
PROTO_PRIM(primGetNextEvt);
PROTO_PRIM(primEvtAvail);

PROTO_PRIM(primCreateCallback);
PROTO_PRIM(primDisposeCallback);
#endif


/* And some "bit-twiddling" primitives, which are generally useful. KH */
#if BIT_OPERATIONS
PROTO_PRIM(primBAnd);
PROTO_PRIM(primBOr);
PROTO_PRIM(primBComp);
PROTO_PRIM(primBASL);
PROTO_PRIM(primBASR);
PROTO_PRIM(primBTst);
PROTO_PRIM(primBSet);
PROTO_PRIM(primBClr);
PROTO_PRIM(primBLSet);
PROTO_PRIM(primBHSet);
#endif

static Void   local printer		Args((StackPtr,Name,Int,Cell));
static Void   local startList		Args((StackPtr,Cell));
static Void   local startNList		Args((StackPtr,Cell));

PROTO_PRIM(primLPrint);
PROTO_PRIM(primNLPrint);
PROTO_PRIM(primSPrint);
PROTO_PRIM(primNSPrint);

static Cell   local textAsVar		Args((Text,Cell));
static Cell   local textAsOp		Args((Text,Cell));
static Cell   local stringOutput	Args((String,Cell));
static Cell   local printBadRedex	Args((Cell,Cell));

static String local evalName		Args((Cell));
static Void   local abandonDialogue	Args((Cell));
static Cell   local printDBadRedex	Args((Cell,Cell));
static Cell   local readFile		Args((Void));
static Cell   local writeFile		Args((Void));
static Cell   local appendFile		Args((Void));
static Cell   local readChan		Args((Void));
static Cell   local appendChan		Args((Void));
static FILE  *local validOutChannel	Args((String));
static Cell   local echo		Args((Void));
static Cell   local getCLArgs		Args((Void));
static Cell   local getProgName		Args((Void));
static Cell   local getEnv		Args((Void));
#if MAC
static Cell   local imperate		Args((Void));
#endif

PROTO_PRIM(primInput);
PROTO_PRIM(primFopen);

#ifdef LAMBDAVAR
PROTO_PRIM(primLvReturn);
PROTO_PRIM(primLvPure);
PROTO_PRIM(primLvRead);
PROTO_PRIM(primLvBind);
PROTO_PRIM(primLvVar);
PROTO_PRIM(primLvNewvar);
PROTO_PRIM(primLvAssign);
PROTO_PRIM(primLvVarEq);
PROTO_PRIM(primLvGetch);
PROTO_PRIM(primLvPutchar);
PROTO_PRIM(primLvSystem);
#endif

#ifdef LAMBDANU
PROTO_PRIM(primLnReturn);
PROTO_PRIM(primLnBind);
PROTO_PRIM(primLnFlip);
PROTO_PRIM(primLnNew);
PROTO_PRIM(primLnAssign);
PROTO_PRIM(primLnRead);
PROTO_PRIM(primLnIo);
PROTO_PRIM(primLnBegin);
PROTO_PRIM(primLnTagEq);
PROTO_PRIM(primLnGetch);
PROTO_PRIM(primLnPutchar);
PROTO_PRIM(primLnSystem);
PROTO_PRIM(primLnDone);
#endif

#endif

/* --------------------------------------------------------------------------
 * Table of primitive/built-in values:
 * ------------------------------------------------------------------------*/

#if PRIMITIVES_CODE
#define GofcPrim(imp)	imp
#define NoGofcPrim(imp)	imp
#else
#define GofcPrim(imp)	PRIM_GOFC
#define NoGofcPrim(imp)	PRIM_NOGOFC
#endif

struct primitive primitives[] = {
  {"primFatbar",	2, GofcPrim(primFatbar)},
  {"primFail",		0, GofcPrim(primFail)},
  {"primUndefMem",	1, GofcPrim(primFail)},
  {"primGCBhole",	0, NoGofcPrim(primFail)},
  {"primError",		1, GofcPrim(primFail)},
  {"primSel",		3, GofcPrim(primSel)},
  {"primIf",		3, GofcPrim(primIf)},
  
  {"primCompare",	1, NoGofcPrim(primCmp)},
  {"primInput",		1, NoGofcPrim(primInput)},
  {"primPrint",		3, NoGofcPrim(primPrint)},
  {"primNprint",	3, NoGofcPrim(primNPrint)},
  {"primLprint",	2, NoGofcPrim(primLPrint)},
  {"primNlprint",	2, NoGofcPrim(primNLPrint)},
  {"primSprint",	2, NoGofcPrim(primSPrint)},
  {"primNsprint",	2, NoGofcPrim(primNSPrint)},
  
  {"primPlusInt",	2, GofcPrim(primPlusInt)},
  {"primMinusInt",	2, GofcPrim(primMinusInt)},
  {"primMulInt",	2, GofcPrim(primMulInt)},
  {"primDivInt",	2, GofcPrim(primDivInt)},
  {"primQuotInt",	2, GofcPrim(primQuotInt)},
  {"primModInt",	2, GofcPrim(primModInt)},
  {"primRemInt",	2, GofcPrim(primRemInt)},
  {"primNegInt",	1, GofcPrim(primNegInt)},
  
  {"primPlusFloat",	2, GofcPrim(primPlusFloat)},
  {"primMinusFloat",	2, GofcPrim(primMinusFloat)},
  {"primMulFloat",	2, GofcPrim(primMulFloat)},
  {"primDivFloat",	2, GofcPrim(primDivFloat)},
  {"primNegFloat",	1, GofcPrim(primNegFloat)},
  
  #if HAS_FLOATS
  {"primSinFloat",	1, GofcPrim(primSinFloat)},
  {"primCosFloat",	1, GofcPrim(primCosFloat)},
  {"primTanFloat",	1, GofcPrim(primTanFloat)},
  {"primAsinFloat",	1, GofcPrim(primAsinFloat)},
  {"primAcosFloat",	1, GofcPrim(primAcosFloat)},
  {"primAtanFloat",	1, GofcPrim(primAtanFloat)},
  {"primAtan2Float",	2, GofcPrim(primAtan2Float)},
  {"primExpFloat",	1, GofcPrim(primExpFloat)},
  {"primLogFloat",	1, GofcPrim(primLogFloat)},
  {"primLog10Float",	1, GofcPrim(primLog10Float)},
  {"primSqrtFloat",	1, GofcPrim(primSqrtFloat)},
  {"primFloatToInt",	1, GofcPrim(primFloatToInt)},

  #if HAS_HYPERBOLICS
  /* We might as well have the Hyperbolic functions -- KH */
  {"primSinhFloat",	1, GofcPrim(primSinhFloat)},
  {"primCoshFloat",	1, GofcPrim(primCoshFloat)},
  {"primTanhFloat",	1, GofcPrim(primTanhFloat)},
  {"primAsinhFloat",	1, GofcPrim(primAsinhFloat)},
  {"primAcoshFloat",	1, GofcPrim(primAcoshFloat)},
  {"primAtanhFloat",	1, GofcPrim(primAtanhFloat)},
  #endif
  #endif
  
  {"primIntToChar",	1, GofcPrim(primIntToChar)},
  {"primCharToInt",	1, GofcPrim(primCharToInt)},
  {"primIntToFloat",	1, GofcPrim(primIntToFloat)},
  
  {"primEqInt",		2, GofcPrim(primEqInt)},
  {"primLeInt",		2, GofcPrim(primLeInt)},
  {"primEqChar",	2, GofcPrim(primEqChar)},
  {"primLeChar",	2, GofcPrim(primLeChar)},
  {"primEqFloat",	2, GofcPrim(primEqFloat)},
  {"primLeFloat",	2, GofcPrim(primLeFloat)},
  
  {"primGenericEq",	2, GofcPrim(primGenericEq)},
  {"primGenericNe",	2, GofcPrim(primGenericNe)},
  {"primGenericGt",	2, GofcPrim(primGenericGt)},
  {"primGenericLe",	2, GofcPrim(primGenericLe)},
  {"primGenericGe",	2, GofcPrim(primGenericGe)},
  {"primGenericLt",	2, GofcPrim(primGenericLt)},
  
  /* C Monad primitives for the Mac -- Gofc versions not yet provided. */
  #if MAC
  {"primUnitIO",          1, NoGofcPrim(primUnitIO)},
  {"primCUnitIO",         2, NoGofcPrim(primCUnitIO)},
  {"primBindIO",          2, NoGofcPrim(primBindIO)},
  {"primCBindIO",         3, NoGofcPrim(primCBindIO)},
  {"primCCBindIO",        2, NoGofcPrim(primCCBindIO)},
 
  {"primTrap",            3, NoGofcPrim(primTrap)},
  {"primTrapReg",         7, NoGofcPrim(primTrapReg)},
  {"primAssign",          2, NoGofcPrim(primAssign)},
  {"primAssignS",         2, NoGofcPrim(primAssignS)},
  {"primAssignC",         2, NoGofcPrim(primAssignC)},
  {"primAssignBlock",     2, NoGofcPrim(primAssignBlock)},
  {"primMalloc",          1, NoGofcPrim(primMalloc)},
  {"primFree",            1, NoGofcPrim(primFree)},
  {"primDeref",           1, NoGofcPrim(primDeref)},
  {"primSeq",             2, NoGofcPrim(primSeq)},
  {"primTrace",           2, NoGofcPrim(primTrace)},

  /* ToolBox calls coded for speed */
  {"primLineTo",          2, NoGofcPrim(primLineTo)},
  {"primMoveTo",          2, NoGofcPrim(primMoveTo)},
  {"primButton",          1, NoGofcPrim(primButton)},
  {"primGetMouse",        1, NoGofcPrim(primGetMouse)},

  /* Event handling variations on standard code */
  {"primGetNextEvent",    1, NoGofcPrim(primGetNextEvt)},
  {"primEventAvail",      1, NoGofcPrim(primEvtAvail)},

  /* Prim */
  {"primCreateCallback",  2, NoGofcPrim(primCreateCallback)},
  {"primDisposeCallback", 1, NoGofcPrim(primDisposeCallback)},
  #endif

  #if BIT_OPERATIONS
  {"primBAnd",        2, GofcPrim(primBAnd)},
  {"primBOr",         2, GofcPrim(primBOr)},
  {"primBComp",       1, GofcPrim(primBComp)},
  {"primBASL",        2, GofcPrim(primBASL)},
  {"primBASR",        2, GofcPrim(primBASR)},
  {"primBTst",        2, GofcPrim(primBTst)},
  {"primBSet",        2, GofcPrim(primBSet)},
  {"primBClr",        2, GofcPrim(primBClr)},
  {"primBLSet",       1, GofcPrim(primBLSet)},
  {"primBHSet",       1, GofcPrim(primBHSet)},
  #endif

  {"primPrint",		3, NoGofcPrim(primPrint)},
  {"primShowsInt",	3, GofcPrim(primPrint)},
  {"primShowsFloat",	3, GofcPrim(primPrint)},
  
  {"primStrict",	2, GofcPrim(primStrict)},
  
  {"primFopen",		3, GofcPrim(primFopen)},

  #ifdef LAMBDAVAR
  {"primLvReturn",	2, NoGofcPrim(primLvReturn)},
  {"primLvPure",	1, NoGofcPrim(primLvPure)},
  {"primLvRead",	3, NoGofcPrim(primLvRead)},
  {"primLvBind",	3, NoGofcPrim(primLvBind)},
  {"primLvVar",		2, NoGofcPrim(primLvVar)},
  {"primLvNewvar",	1, NoGofcPrim(primLvNewvar)},
  {"primLvAssign",	3, NoGofcPrim(primLvAssign)},
  {"primLvVarEq",	2, NoGofcPrim(primLvVarEq)},
  {"primLvUnbound",	0, NoGofcPrim(primFail)},
  {"primLvGetch",	1, NoGofcPrim(primLvGetch)},
  {"primLvPutchar",	2, NoGofcPrim(primLvPutchar)},
  {"primLvSystem",	2, NoGofcPrim(primLvSystem)},
  #endif
  
#ifdef LAMBDANU
  {"primLnReturn",	2, NoGofcPrim(primLnReturn)},
  {"primLnBind",	3, NoGofcPrim(primLnBind)},
  {"primLnFlip",	3, NoGofcPrim(primLnFlip)},
  {"primLnNew",		1, NoGofcPrim(primLnNew)},
  {"primLnAssign",	3, NoGofcPrim(primLnAssign)},
  {"primLnRead",	3, NoGofcPrim(primLnRead)},
  {"primLnIo",		2, NoGofcPrim(primLnIo)},
  {"primLnBegin",	1, NoGofcPrim(primLnBegin)},
  {"primLnTagEq",	2, NoGofcPrim(primLnTagEq)},
  {"primLnGetch",	1, NoGofcPrim(primLnGetch)},
  {"primLnPutchar",	2, NoGofcPrim(primLnPutchar)},
  {"primLnSystem",	2, NoGofcPrim(primLnSystem)},
  {"primLnUnbound",	0, NoGofcPrim(primFail)},
  {"primLnNocont",	0, NoGofcPrim(primFail)},
  {"primLnDone",	1, NoGofcPrim(primLnDone)},
#endif

  {0,			0, 0}
};

#if 0
dotrace(s,e)
String s;
Cell e;
{
  fprintf(stderr,"%s",s);
  printExp(stderr,e);
  fputc('\n',stderr);
}
#else
#define dotrace(s,e)
#endif

/* --------------------------------------------------------------------------
 * Primitive functions:
 * ------------------------------------------------------------------------*/

#if PRIMITIVES_CODE
primFun(primFatbar) {			/* Fatbar primitive		   */
    Cell l    = primArg(2);		/* _FAIL [] r = r		   */
    Cell r    = primArg(1);		/* l     [] r = l  -- otherwise	   */
    Cell temp = evalWithNoError(l);
    if (nonNull(temp))
	if (temp==nameFail)
	    updateRoot(r);
	else {
	    updateRoot(temp);
	    cantReduce();
	}
    else
	updateRoot(l);
}

primFun(primFail) {		       /* Failure primitive		   */
    cantReduce();
}

primFun(primSel) {		       /* Component selection		   */
    Cell c = primArg(3);	       /* _sel c e n   return nth component*/
    Cell e = primArg(2);	       /*	       in expression e	   */
    Cell n = intOf(primArg(1));        /*	       built using cfun c  */

    eval(e);
    if (whnfHead==c &&	((isName(whnfHead) && name(whnfHead).arity==whnfArgs)
		      || (isTuple(whnfHead) && tupleOf(whnfHead)==whnfArgs)))
	updateRoot(pushed(n-1));
    else
	cantReduce();
}

primFun(primIf) {		       /* Conditional primitive 	   */
    eval(primArg(3));
    if (whnfHead==nameTrue)
	updateRoot(primArg(2));
    else
	updateRoot(primArg(1));
}

primFun(primStrict) {		       /* Strict application primitive	   */
    eval(primArg(1));		       /* evaluate 2nd argument 	   */
    updapRoot(primArg(2),primArg(1));  /* and apply 1st argument to result */
}

/* --------------------------------------------------------------------------
 * Integer arithmetic primitives:
 * ------------------------------------------------------------------------*/

primFun(primPlusInt) {		       /* Integer addition primitive	   */
    Int x;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));
    updateRoot(mkInt(x+whnfInt));
}

primFun(primMinusInt) { 	       /* Integer subtraction primitive    */
    Int x;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));
    updateRoot(mkInt(x-whnfInt));
}

primFun(primMulInt) {		       /* Integer multiplication primitive */
    Int x;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));
    updateRoot(mkInt(x*whnfInt));
}

primFun(primQuotInt) {			/* Integer division primitive	   */
    Int x;				/* truncated towards zero	   */
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));

    if (whnfInt==0)
	cantReduce();

    updateRoot(mkInt(x/whnfInt));
}

primFun(primDivInt) {			/* Integer division primitive	   */
    Int x,r;				/* truncated towards -ve infinity  */
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));

    if (whnfInt==0)
	cantReduce();
    r = x%whnfInt;
    x = x/whnfInt;
    if ((whnfInt<0 && r>0) || (whnfInt>0 && r<0))
	x--;
    updateRoot(mkInt(x));
}

/*
	Isn't x%y undefined if either x or y is negative,
	at least in original K&R C?  
	These two definitions may only work for an ANSI C compiler,
	in fact they seem to have broken for MPW C in the past...

	KH
*/

primFun(primModInt) {		       /* Integer modulo primitive	   */
    Int x,y;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));
    if (whnfInt==0)
	cantReduce();
    y = x%whnfInt;		       /* "... the modulo having the sign  */
    if ((y<0 && whnfInt>0) ||	       /*	       of the divisor ..." */
	(y>0 && whnfInt<0))	       /* See definition on p.91 of Haskell*/
	updateRoot(mkInt(y+whnfInt));  /* report...			   */
    else
	updateRoot(mkInt(y));
}

primFun(primRemInt) {		       /* Integer remainder primitive	   */
    Int x;
    eval(primArg(2));		       /* div and rem satisfy:		   */
    x = whnfInt;		       /* (x `div` y)*y + (x `rem` y) == x */
    eval(primArg(1));		       /* which is exactly the property    */
    if (whnfInt==0)		       /* described in K&R 2:		   */
	cantReduce();		       /*      (a/b)*b + a%b == a	   */
    updateRoot(mkInt(x%whnfInt));
}

primFun(primNegInt) {		       /* Integer negation primitive	   */
    eval(primArg(1));
    updateRoot(mkInt(-whnfInt));
}

/* --------------------------------------------------------------------------
 * Coercion primitives:
 * ------------------------------------------------------------------------*/

primFun(primCharToInt) {	       /* Character to integer primitive   */
    eval(primArg(1));
    updateRoot(mkInt(charOf(whnfHead)));
}

primFun(primIntToChar) {	       /* Integer to character primitive   */
    eval(primArg(1));
    if (whnfInt<0  || whnfInt>MAXCHARVAL)
	cantReduce();
    updateRoot(mkChar(whnfInt));
}

#if BIT_OPERATIONS
primFun(primBAnd) {			/* Bytewise And primitive	   */
    unsigned x;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));

    updateRoot(mkInt((unsigned)(x&(unsigned)whnfInt)));
}

primFun(primBOr) {			/* Bytewise Or primitive	   */
    unsigned x;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));

    updateRoot(mkInt((unsigned)(x|(unsigned)whnfInt)));
}

primFun(primBComp) {			/* Bytewise Complement primitive   */
    eval(primArg(1));

    updateRoot(mkInt((unsigned)(~(unsigned)whnfInt)));
}

primFun(primBASL) {			/* Arithmetic Shift Left primitive   */
    unsigned x;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));

    updateRoot(mkInt((unsigned)(x<<(unsigned)whnfInt)));
}

primFun(primBASR) {			/* Arithmetic Shift Right primitive   */
    unsigned x;				/* To avoid sign-extension	      */
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));

    updateRoot(mkInt((unsigned)(x>>(unsigned)whnfInt)));
}

primFun(primBTst) {			/* Bit Testing primitive	   */
    unsigned x, b;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));

    b = 1 << (whnfInt-1);
    updateRoot((x&b)==b ? nameTrue: nameFalse);
}

primFun(primBSet) {			/* Bit Setting primitive	   */
    Int x;
    Int b;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));

    b = 1 << (whnfInt-1);
    updateRoot(mkInt(x|b));
}

primFun(primBClr) {			/* Bit Clearing primitive	   */
    Int x;
    Int b;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));

    b = 1 << (whnfInt-1);
    updateRoot(mkInt(x&~b));
}


primFun(primBLSet) {			/* Locate least set bit	   */
    Int b;
    unsigned mask;
    eval(primArg(1));
    
    for(b=1,mask=1;mask!=0;++b,mask<<=1)
      if(whnfInt&mask)
        break;

    if(mask==0)
      b=0;
    updateRoot(mkInt(b));
}

/*
	This algorithm works from the lowest bit, so that it
	works on ANY word-size machine without modification.
	This makes it rather inefficient.  Better suggestions
	are welcome!  KH
*/

primFun(primBHSet) {			/* Locate highest set bit	   */
    Int b;
    Int highest;
    unsigned mask;

    eval(primArg(1));
    
    for(b=highest=1,mask=1;mask!=0;++b,mask<<=1)
      if(whnfInt&mask)
        highest=b;

    updateRoot(mkInt(highest));
}

#endif

#if MPW
/*
   These hacks are needed to get around the garbage collector
   which interferes with float passing conventions on the Mac.
   See machine.c for the full details -- KH
*/

union {Float c; float f;} fi, gi;

UPDFLOAT(root,result,f)
Cell root, result;
float f;
{
    fi.f = f;
    snd(result) = fi.c;
    updateRoot(result);
}

Int FLZERO()
{
    fi.c = whnfFloat;
    return(fi.f==0.0);
}

Int EQFLOAT(x)
Float x;
{
   fi.c = whnfFloat;
   gi.c = x;
   return(gi.f==fi.f);
}

Int LEFLOAT(x)
Float x;
{
   fi.c = whnfFloat;
   gi.c = x;
   return(gi.f<=fi.f);
}

Int GTFLOAT(w)
{
  fi.c = whnfFloat;
  gi.c = floatOf(w);
  return(gi.f > fi.f);
}

Int LTFLOAT(w)
{
  fi.c = whnfFloat;
  gi.c = floatOf(w);
  return(gi.f < fi.f);
}

#else
#define UPDFLOAT(r,w)	updateRoot(mkFloat((Float)(w)))
#define GTFLOAT(w)	floatOf(w) > whnfFloat
#define LTFLOAT(w)	floatOf(w) < whnfFloat
#define FNEGATIVE(w)	floatOf(w) < 0.0
#endif


#if !MPW
primFun(primIntToFloat) {		/* Integer to Float primitive	   */
    eval(primArg(1));
    updateRoot(mkFloat((Float)(whnfInt)));
}

/* --------------------------------------------------------------------------
 * Float arithmetic primitives:
 * ------------------------------------------------------------------------*/

primFun(primPlusFloat) {	       /* Float addition primitive	   */
    Float x;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    updateRoot(mkFloat(x+whnfFloat));
}

primFun(primMinusFloat) { 	       /* Float subtraction primitive	   */
    Float x;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    updateRoot(mkFloat(x-whnfFloat));
}

primFun(primMulFloat) {		       /* Float multiplication primitive   */
    Float x;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    updateRoot(mkFloat(x*whnfFloat));
}

primFun(primDivFloat) {		       /* Float division primitive	   */
    Float x;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    if (whnfFloat==0)
	cantReduce();
    updateRoot(mkFloat(x/whnfFloat));
}

primFun(primNegFloat) {		       /* Float negation primitive	   */
    eval(primArg(1));
    updateRoot(mkFloat(-whnfFloat));
}

#if HAS_FLOATS
primFun(primSinFloat) {			/* Float sin (trig) primitive	   */
    eval(primArg(1));
    updateRoot(mkFloat(sin(whnfFloat)));
}

primFun(primCosFloat) {			/* Float cos (trig) primitive	   */
    eval(primArg(1));
    updateRoot(mkFloat(cos(whnfFloat)));
}

primFun(primTanFloat) {			/* Float tan (trig) primitive	   */
    eval(primArg(1));
    updateRoot(mkFloat(tan(whnfFloat)));
}

primFun(primAsinFloat) {		/* Float arc sin (trig) primitive  */
    eval(primArg(1));
    updateRoot(mkFloat(asin(whnfFloat)));
}

primFun(primAcosFloat) {		/* Float arc cos (trig) primitive  */
    eval(primArg(1));
    updateRoot(mkFloat(acos(whnfFloat)));
}

primFun(primAtanFloat) {		/* Float arc tan (trig) primitive  */
    eval(primArg(1));
    updateRoot(mkFloat(atan(whnfFloat)));
}

#if HAS_HYPERBOLICS
primFun(primSinhFloat) {			/* Hyperbolic Float sin (trig) primitive	   */
    eval(primArg(1));
    updateRoot(mkFloat(sinh(whnfFloat)));
}

primFun(primCoshFloat) {			/* Hyperbolic Float cos (trig) primitive	   */
    eval(primArg(1));
    updateRoot(mkFloat(cosh(whnfFloat)));
}

primFun(primTanhFloat) {			/* Hyperbolic Float tan (trig) primitive	   */
    eval(primArg(1));
    updateRoot(mkFloat(tanh(whnfFloat)));
}

primFun(primAsinhFloat) {		/* Hyperbolic Float arc sin (trig) primitive  */
    eval(primArg(1));
    updateRoot(mkFloat(asinh(whnfFloat)));
}

primFun(primAcoshFloat) {		/* Hyperbolic Float arc cos (trig) primitive  */
    eval(primArg(1));
    updateRoot(mkFloat(acosh(whnfFloat)));
}

primFun(primAtanhFloat) {		/* Hyperbolic Float arc tan (trig) primitive  */
    eval(primArg(1));
    updateRoot(mkFloat(atanh(whnfFloat)));
}
#endif

primFun(primAtan2Float) {		/* Float arc tan with quadrant info*/
    Float t;				/* 		 (trig) primitive  */
    eval(primArg(2));
    t = whnfFloat;
    eval(primArg(1));
    updateRoot(mkFloat(atan2(t,whnfFloat)));
}

primFun(primExpFloat) {			/* Float exponential primitive	   */
    eval(primArg(1));
    updateRoot(mkFloat(exp(whnfFloat)));
}

primFun(primLogFloat) {			/* Float logarithm primitive	   */
    eval(primArg(1));
    if (whnfFloat<=0)
	cantReduce();
    updateRoot(mkFloat(log(whnfFloat)));
}

/* ??why is this primitive?? KH: log10(x) = log(x)/log(10) */
primFun(primLog10Float) {		/* Float logarithm (base 10) prim  */
    eval(primArg(1));
    if (whnfFloat<=0)
	cantReduce();
    updateRoot(mkFloat(log10(whnfFloat)));
}

primFun(primSqrtFloat) {		/* Float square root primitive	   */
    eval(primArg(1));
    if (whnfFloat<0)
	cantReduce();
    updateRoot(mkFloat(sqrt(whnfFloat)));
}

primFun(primFloatToInt) {		/* Adhoc Float --> Int conversion  */
    eval(primArg(1));

/*
   My version is probably better for negative floats 
   -- assuming you want truncation...  KH
*/
#if 0
    updateRoot(mkInt((Int)(whnfFloat)));
#else
    {
      Int ftoi = (Int) whnfFloat;
      updateRoot(mkInt(whnfFloat>=0.0? ftoi:
                       (Float) ftoi == whnfFloat? ftoi:
	               (ftoi-1)));
    }
#endif
}
#endif


#else	/* !MPW */

#pragma segment Builtin2

#define createFloatResult()	mkFloat(0)

Boolean FNEGATIVE(x)
Float x;
{
    fi.c = x;
    return(fi.f<0.0);
}


ITOF(root,result,w)
Cell root, result;
Int w;
{
    fi.f = (float) w;
    UPDFLOAT(root,result,fi.f);
}

FTOI(root)
Cell root;
{
    fi.c = whnfFloat;
    updateRoot(mkInt(fi.f>=0.0?(int)fi.f:
                     (float)((int)(fi.f))==fi.f?(int)fi.f:
	              ((int)fi.f)-1));
}

ADDFLOAT(root,result,x)
Cell root,result;
Float x;
{
   fi.c = x;
   gi.c = whnfFloat;
   UPDFLOAT(root,result,fi.f+gi.f);
}

SUBFLOAT(root,result,x)
Cell root,result;
Float x;
{
   fi.c = x;
   gi.c = whnfFloat;
   UPDFLOAT(root,result,fi.f-gi.f);
}

MULFLOAT(root,result,x)
Cell root,result;
Float x;
{
   fi.c = x;
   gi.c = whnfFloat;
   UPDFLOAT(root,result,fi.f*gi.f);
}

DIVFLOAT(root,result,x)
Cell root, result;
Float x;
{
   fi.c = x;
   gi.c = whnfFloat;
   UPDFLOAT(root,result,fi.f/gi.f);
}

NEGFLOAT(root,result)
Cell root, result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,-gi.f);
}

SQRT(root,result)
Cell root, result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,sqrt(gi.f));
}

LOG(root,result)
Cell root, result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,log(gi.f));
}

LOG10(root,result)
Cell root, result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,log10(gi.f));
}

EXP(root,result)
Cell root, result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,exp(gi.f));
}

SIN(root,result)
Cell root,result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,sin(gi.f));
}

COS(root,result)
Cell root,result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,cos(gi.f));
}

TAN(root,result)
Cell root,result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,tan(gi.f));
}

ASIN(root,result)
Cell root,result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,asin(gi.f));
}

ACOS(root,result)
Cell root,result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,acos(gi.f));
}

ATAN(root,result)
Cell root,result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,atan(gi.f));
}

ATAN2(root,result,x)
Cell root,result;
Float x;
{
   fi.c = x;
   gi.c = whnfFloat;
   UPDFLOAT(root,result,atan2(fi.f,gi.f));
}

SINH(root,result)
Cell root,result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,sinh(gi.f));
}

COSH(root,result)
Cell root,result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,cosh(gi.f));
}

TANH(root,result)
Cell root,result;
{
   gi.c = whnfFloat;
   UPDFLOAT(root,result,tanh(gi.f));
}

ASINH(root,result)
Cell root,result;
{
#if 1
   updapRoot(nameAsinh,whnfFloat);
#else
   gi.c = whnfFloat;
   UPDFLOAT(root,result,asinh(gi.f));
#endif
}

ACOSH(root,result)
Cell root,result;
{
#if 1
   updapRoot(nameAcosh,whnfFloat);
#else
   gi.c = whnfFloat;
   UPDFLOAT(root,result,acosh(gi.f));
#endif
}

ATANH(root,result)
Cell root,result;
{
#if 1
   updapRoot(nameAtanh,whnfFloat);
#else
   gi.c = whnfFloat;
   UPDFLOAT(root,result,atanh(gi.f));
#endif
}

primFun(primIntToFloat) {		/* Integer to Float primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    ITOF(root,result,whnfInt);
}

primFun(primPlusFloat) {	       /* Float addition primitive	   */
    Float x, result = 0;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    result = createFloatResult();
    ADDFLOAT(root,result,x);
}

primFun(primMinusFloat) { 	       /* Float subtraction primitive	   */
    Float x, result = 0;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    result = createFloatResult();
    SUBFLOAT(root,result,x);
}

primFun(primMulFloat) {		       /* Float multiplication primitive   */
    Float x, result = 0;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    result = createFloatResult();
    MULFLOAT(root,result,x);
}


primFun(primDivFloat) {		       /* Float division primitive	   */
    Float x, result = 0;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    if (FLZERO())
	cantReduce();
    result = createFloatResult();
    DIVFLOAT(root,result,x);
}

primFun(primNegFloat) {		       /* Float negation primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    NEGFLOAT(root,result);
}

primFun(primFloatToInt) {		/* Float Truncation primitive	   */
    eval(primArg(1));
    FTOI(root);
}

primFun(primSqrtFloat) {		/* Square Root primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    SQRT(root,result);
}

primFun(primLogFloat) {			/* Natural Logarithm primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    LOG(root,result);
}

primFun(primLog10Float) {		/* Logarithm (base 10) primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    LOG10(root,result);
}

primFun(primExpFloat) {			/* Inverse Logarithm primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    EXP(root,result);
}

primFun(primSinFloat) {			/* Sine primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    SIN(root,result);
}

primFun(primCosFloat) {			/* Cosine primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    COS(root,result);
}

primFun(primTanFloat) {			/* Tangent primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    TAN(root,result);
}

primFun(primAsinFloat) {		/* ArcSin primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    ASIN(root,result);
}

primFun(primAcosFloat) {		/* ArcCos primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    ACOS(root,result);
}

primFun(primAtanFloat) {		/* ArcTangent primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    ATAN(root,result);
}

primFun(primAtan2Float) {		/* Float arc tan with quadrant info*/
    Float x, result = 0;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    result = createFloatResult();
    ATAN2(root,result,x);
}

#if HAS_HYPERBOLICS
primFun(primSinhFloat) {		/* Hyperbolic Sine primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    SINH(root,result);
}

primFun(primCoshFloat) {		/* Hyperbolic Cosine primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    COSH(root,result);
}

primFun(primTanhFloat) {		/* Hyperbolic Tangent primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    TANH(root,result);
}

primFun(primAsinhFloat) {		/* Hyperbolic ArcSin primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    ASINH(root,result);
}

primFun(primAcoshFloat) {		/* Hyperbolic ArcCos primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    ACOSH(root,result);
}

primFun(primAtanhFloat) {		/* Hyperbolic ArcTangent primitive	   */
    Float result = 0;
    eval(primArg(1));
    result = createFloatResult();
    ATANH(root,result);
}
#endif
#endif


/* --------------------------------------------------------------------------
 * Comparison primitives:
 * ------------------------------------------------------------------------*/

primFun(primEqInt) {		       /* Integer equality primitive	   */
    Int x;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));
    updateRoot(x==whnfInt ? nameTrue : nameFalse);
}

primFun(primLeInt) {		       /* Integer <= primitive		   */
    Int x;
    eval(primArg(2));
    x = whnfInt;
    eval(primArg(1));
    updateRoot(x<=whnfInt ? nameTrue : nameFalse);
}

primFun(primEqChar) {		       /* Character equality primitive	   */
    Cell x;
    eval(primArg(2));
    x = whnfHead;
    eval(primArg(1));
    updateRoot(x==whnfHead ? nameTrue : nameFalse);
}

primFun(primLeChar) {		       /* Character <= primitive	   */
    Cell x;
    eval(primArg(2));
    x = whnfHead;
    eval(primArg(1));
    updateRoot(x<=whnfHead ? nameTrue : nameFalse);
}

#if !MPW
primFun(primEqFloat) {		       /* Float equality primitive	   */
    Float x;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    updateRoot(x==whnfFloat ? nameTrue : nameFalse);
}

primFun(primLeFloat) {		       /* Float <= primitive		   */
    Float x;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    updateRoot(x<=whnfFloat ? nameTrue : nameFalse);
}
#else	/* !MPW */
primFun(primEqFloat) {		       /* Float equality primitive	   */
    Float x;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    updateRoot(EQFLOAT(x) ? nameTrue : nameFalse);
}

primFun(primLeFloat) {		       /* Float equality primitive	   */
    Float x;
    eval(primArg(2));
    x = whnfFloat;
    eval(primArg(1));
    updateRoot(LEFLOAT(x) ? nameTrue : nameFalse);
}
#endif


#if MPW
#pragma segment PrimCmp
#endif

/* Generic comparisons implemented using the internal primitive function:
 *
 * primCmp []			= EQ
 *         ((C xs, D ys):rs)
 *	   | C < D		= LT
 *	   | C == D		= primCmp (zip xs ys ++ rs)
 *	   | C > D		= GT
 *	   ((Int n, Int m):rs)
 *	   | n < m		= LT
 *	   | n == m		= primCmp rs
 *	   | n > m		= GT
 *	   etc ... similar for comparison of characters:
 *
 * The list argument to primCmp is represented as an `internal list';
 * i.e. no (:)/[] constructors - use internal cons and NIL instead!
 *
 * To compare two values x and y, evaluate primCmp [(x,y)] and use result.
 */

#define LT            1
#define EQ            2
#define GT            3
#define compResult(x) updateRoot(mkInt(x))

static Name namePrimCmp;

primFun(primCmp) {			/* generic comparison function	   */
    Cell rs = primArg(1);

    if (isNull(rs)) {
	compResult(EQ);
	return;
    }
    else {
	Cell x = fst(hd(rs));
	Cell y = snd(hd(rs));
	Int  whnfArgs1;
	Cell whnfHead1;

	rs = tl(rs);
	eval(x);
	whnfArgs1 = whnfArgs;
	whnfHead1 = whnfHead;

	switch (whatIs(whnfHead1)) {
	    case INTCELL  : if (whnfArgs==0) {		/* compare ints    */
				eval(y);
				if (!isInt(whnfHead) || whnfArgs!=0)
				    break;
				if (intOf(whnfHead1) > whnfInt)
				    compResult(GT);
				else if (intOf(whnfHead1) < whnfInt)
				    compResult(LT);
				else
				    updapRoot(namePrimCmp,rs);
				return;
			    }
			    break;

	    case FLOATCELL: if (whnfArgs==0) {		/* compare floats  */
				eval(y);
				if (!isFloat(whnfHead) || whnfArgs!=0)
				    break;
				if (GTFLOAT(whnfHead1))
				    compResult(GT);
				else if (LTFLOAT(whnfHead1))
				    compResult(LT);
				else
				    updapRoot(namePrimCmp,rs);
				return;
			    }
			    break;

	    case CHARCELL : if (whnfArgs==0) {		/* compare chars   */
				eval(y);
				if (!isChar(whnfHead) || whnfArgs!=0)
				    break;
				if (charOf(whnfHead1) > charOf(whnfHead))
				    compResult(GT);
				else if (charOf(whnfHead1) < charOf(whnfHead))
				    compResult(LT);
				else
				    updapRoot(namePrimCmp,rs);
				return;
			    }
			    break;

	    default	  : eval(y);			/* compare structs */
			    if (whnfHead1==whnfHead &&
				whnfArgs1==whnfArgs &&
				(whnfHead==UNIT    ||
				 isTuple(whnfHead) ||
				 (isName(whnfHead) &&
				  name(whnfHead).defn==CFUN))) {
				while (whnfArgs1-- >0)
				    rs = cons(pair(pushed(whnfArgs+whnfArgs1),
						   pushed(whnfArgs1)),rs);
				updapRoot(namePrimCmp,rs);
				return;
			    }
			    if (isName(whnfHead1)	    &&
				 name(whnfHead1).defn==CFUN &&
				 isName(whnfHead)	    &&
				 name(whnfHead).defn==CFUN) {
				if (name(whnfHead1).number
						> name(whnfHead).number)
				    compResult(GT);
				else if (name(whnfHead1).number
						< name(whnfHead).number)
				    compResult(LT);
				else
				    break;
				return;
			    }
                            break;
	}
        /* we're going to fail because we can't compare x and y; modify    */
	/* the root expression so that it looks reasonable before failing  */
	/* i.e. output produced will be:  {_compare x y}		   */
	updapRoot(ap(namePrimCmp,x),y);
    }
    cantReduce();
}

primFun(primGenericEq) {		/* Generic equality test	   */
    Cell c = ap(namePrimCmp,singleton(pair(primArg(2),primArg(1))));
    eval(c);
    updateRoot(whnfInt==EQ ? nameTrue : nameFalse);
}

primFun(primGenericLe) {		/* Generic <= test		   */
    Cell c = ap(namePrimCmp,singleton(pair(primArg(2),primArg(1))));
    eval(c);
    updateRoot(whnfInt<=EQ ? nameTrue : nameFalse);
}

primFun(primGenericLt) {		/* Generic < test		   */
    Cell c = ap(namePrimCmp,singleton(pair(primArg(2),primArg(1))));
    eval(c);
    updateRoot(whnfInt<EQ ? nameTrue : nameFalse);
}

primFun(primGenericGe) {		/* Generic >= test		   */
    Cell c = ap(namePrimCmp,singleton(pair(primArg(2),primArg(1))));
    eval(c);
    updateRoot(whnfInt>=EQ ? nameTrue : nameFalse);
}

primFun(primGenericGt) {		/* Generic > test		   */
    Cell c = ap(namePrimCmp,singleton(pair(primArg(2),primArg(1))));
    eval(c);
    updateRoot(whnfInt>EQ ? nameTrue : nameFalse);
}

primFun(primGenericNe) {		/* Generic /= test		   */
    Cell c = ap(namePrimCmp,singleton(pair(primArg(2),primArg(1))));
    eval(c);
    updateRoot(whnfInt!=EQ ? nameTrue : nameFalse);
}

/* --------------------------------------------------------------------------
 * Print primitives:
 * ------------------------------------------------------------------------*/

#if MPW
#pragma segment PrimMisc
#endif

static Cell consOpen,	consSpace,  consComma,	consClose;
static Cell consObrace, consCbrace, consOsq,	consCsq;
static Cell consBack,	consMinus,  consQuote,  consDQuote;

static Name nameLPrint, nameNLPrint;	/* list printing primitives	   */
static Name nameSPrint, nameNSPrint;	/* string printing primitives	   */

#define print(pr,d,e,ss)    ap(ap(ap(pr,mkInt(d)),e),ss)
#define lprint(pr,xs,ss)    ap(ap(pr,xs),ss)
#define printString(s,ss)   revOnto(stringOutput(s,NIL),ss)
#define printSChar(c,ss)    printString(unlexChar(c,'\"'),ss)

primFun(primPrint) {			/* evaluate and print term	   */
    Int  d    = intOf(primArg(3));	/*    :: Int->Expr->[Char]->[Char] */
    Cell e    = primArg(2);
    Cell ss   = primArg(1);
    Cell temp = evalWithNoError(e);
    if (nonNull(temp))
	updateRoot(printBadRedex(temp,ss));
    else
	printer(root,namePrint,d,ss);
}

primFun(primNPrint) {			/* print term without evaluation   */
    Int    d	  = intOf(primArg(3)); /*     :: Int->Expr->[Char]->[Char] */
    Cell   e	  = primArg(2);
    Cell   ss	  = primArg(1);
    unwind(e);
    printer(root,nameNPrint,d,ss);
}

static Void local printer(root,pr,d,ss)	/* Main part: primPrint/primNPrint */
StackPtr root;				/* root or print redex		   */
Name	 pr;				/* printer to use on components	   */
Int	 d;				/* precedence level		   */
Cell	 ss; {				/* rest of output		   */
    Int  used	= 0;
    Cell output = NIL;

    switch(whatIs(whnfHead)) {

	case NAME     : {   Syntax sy = syntaxOf(name(whnfHead).text);

			    if (name(whnfHead).defn!=CFUN ||
				    name(whnfHead).arity>whnfArgs)
				pr = nameNPrint;

			    if (whnfHead==nameCons && whnfArgs==2) {/*list */
				if (pr==namePrint)
				    startList(root,ss);
				else
				    startNList(root,ss);
				return;
			    }
			    if (whnfArgs==1 && sy!=APPLIC) {	  /* (e1+) */
				used   = 1;
				output = ap(consClose,
					  textAsOp(name(whnfHead).text,
					   ap(consSpace,
					    print(pr,FUN_PREC-1,pushed(0),
					     ap(consOpen,NIL)))));
			    }
			    else if (whnfArgs>=2 && sy!=APPLIC) { /* e1+e2 */
				Syntax a = assocOf(sy);
				Int    p = precOf(sy);
				used     = 2;
				if (whnfArgs>2 || d>p)
				     output = ap(consOpen,output);
				output = print(pr,(a==RIGHT_ASS?p:1+p),
					      pushed(1),
					  ap(consSpace,
					   textAsOp(name(whnfHead).text,
					    ap(consSpace,
					     print(pr,(a==LEFT_ASS? p:1+p),
						  pushed(0),
					      output)))));
				if (whnfArgs>2 || d>p)
				    output = ap(consClose,output);
			    }
			    else				  /* f ... */
				output = textAsVar(name(whnfHead).text,NIL);
			}
			break;

	case INTCELL  : {   Int digit;

			    if (intOf(whnfHead)<0 && d>=FUN_PREC)
				output = ap(consClose,output);

			    do {
				digit = whnfInt%10;
				if (digit<0)
				    digit= (-digit);
				output = ap(consChar('0'+digit),output);
			    } while ((whnfInt/=10)!=0);

			    if (intOf(whnfHead)<0) {
				output = ap(consMinus,output);
				if (d>=FUN_PREC)
				    output = ap(consOpen,output);
			    }

			    output = rev(output);
			    pr	   = nameNPrint;
			}
			break;

	case UNIT     : output = ap(consClose,ap(consOpen,NIL));
			pr     = nameNPrint;
			break;

	case TUPLE    : {   Int  tn   = tupleOf(whnfHead);
                            Cell punc = consOpen;
			    Int  i;

			    used      = tn<whnfArgs ? tn : whnfArgs;
			    output    = NIL;
			    for (i=0; i<used; ++i) {
				output = print(pr,MIN_PREC,pushed(i),
					  ap(punc,
					   output));
				punc   = consComma;
			    }
			    for (; i<tn; ++i) {
				output = ap(punc,output);
				punc   = consComma;
			    }
			    output = ap(consClose,output);
			}
			pr = nameNPrint;
			break;

	case CHARCELL : output = ap(consQuote,
                                  stringOutput(unlexChar(charOf(whnfHead),
                                                         '\''),
				   ap(consQuote,
				    output)));
			pr     = nameNPrint;
			break;

	case FLOATCELL:
	
/* The standard Gofer was broken here -- needs parens around negative floats...  KH */ 
			if (FNEGATIVE(whnfFloat) && d>=FUN_PREC)
		          output = ap(consOpen,output);
	                output = stringOutput(floatToString(whnfFloat),
					      output);
		        if (FNEGATIVE(whnfFloat) && d>=FUN_PREC)
		          output = ap(consClose,output);
			pr     = nameNPrint;
			break;

        case DICTCELL : output = stringOutput("{dict}",output);
			pr     = nameNPrint;
			break;

	case FILECELL : output = stringOutput("{file}",output);
			pr     = nameNPrint;
			break;

	default       : internal("Error in graph");
			break;
    }

    if (used<whnfArgs) {		/* Add remaining args to output	   */
	do
	    output = print(pr,FUN_PREC,pushed(used),ap(consSpace,output));
	while (++used<whnfArgs);

	if (d>=FUN_PREC) {		/* Determine if parens are needed  */
	    updapRoot(consOpen,revOnto(output,ap(consClose,ss)));
	    return;
	}
    }

    updateRoot(revOnto(output,ss));
}

/* --------------------------------------------------------------------------
 * List printing primitives:
 * ------------------------------------------------------------------------*/

static Void local startList(root,ss)	/* start printing evaluated list   */
StackPtr root;
Cell     ss; {
    Cell x    = pushed(0);
    Cell xs   = pushed(1);
    Cell temp = evalWithNoError(x);
    if (nonNull(temp))
	updapRoot(consOsq,
		   printBadRedex(temp,
		    lprint(nameLPrint,xs,ss)));
    else if (isChar(whnfHead) && whnfArgs==0)
	updapRoot(consDQuote,
		   printSChar(charOf(whnfHead),
		    lprint(nameSPrint,xs,ss)));
    else
	updapRoot(consOsq,
		   print(namePrint,MIN_PREC,x,
		    lprint(nameLPrint,xs,ss)));
}

static Void local startNList(root,ss)	/* start printing unevaluated list */
StackPtr root;
Cell     ss; {
    Cell x    = pushed(0);
    Cell xs   = pushed(1);
    unwind(x);
    if (isChar(whnfHead) && whnfArgs==0)
	updapRoot(consDQuote,
		   printSChar(charOf(whnfHead),
		    lprint(nameNSPrint,xs,ss)));
    else
	updapRoot(consOsq,
		   print(nameNPrint,MIN_PREC,x,
		    lprint(nameNLPrint,xs,ss)));
}

primFun(primLPrint) {			/* evaluate and print list	   */
    Cell e    = primArg(2);
    Cell ss   = primArg(1);
    Cell temp = evalWithNoError(e);

    if (nonNull(temp))
	updateRoot(printString("] ++ ",printBadRedex(temp,ss)));
    else if (whnfHead==nameCons && whnfArgs==2)
	updapRoot(consComma,
		   ap(consSpace,
		    print(namePrint,MIN_PREC,pushed(0),
		     lprint(nameLPrint,pushed(1),ss))));
    else if (whnfHead==nameNil && whnfArgs==0)
	updapRoot(consCsq,ss);
    else
	updateRoot(printString("] ++ ",printBadRedex(e,ss)));
}

primFun(primNLPrint) {			/* print list without evaluation   */
    Cell e  = primArg(2);
    Cell ss = primArg(1);
    unwind(e);
    if (whnfHead==nameCons && whnfArgs==2)
	updapRoot(consComma,
		   ap(consSpace,
		    print(nameNPrint,MIN_PREC,pushed(0),
		     lprint(nameNLPrint,pushed(1),ss))));
    else if (whnfHead==nameNil && whnfArgs==0)
	updapRoot(consCsq,ss);
    else
	updateRoot(printString("] ++ ",print(nameNPrint,FUN_PREC-1,e,ss)));
}

primFun(primSPrint) {			/* evaluate and print string	   */
    Cell e    = primArg(2);
    Cell ss   = primArg(1);
    Cell temp = evalWithNoError(e);

    if (nonNull(temp))
	updateRoot(printString("\" ++ ",printBadRedex(temp,ss)));
    else if (whnfHead==nameCons && whnfArgs==2) {
	Cell x  = pushed(0);
	Cell xs = pushed(1);
	temp    = evalWithNoError(x);
	if (nonNull(temp))
	    updateRoot(printString("\" ++ [",
			printBadRedex(temp,
			 lprint(nameLPrint,xs,ss))));
	else if (isChar(whnfHead) && whnfArgs==0)
	    updateRoot(printSChar(charOf(whnfHead),
		        lprint(nameSPrint,xs,ss)));
	else
	    updateRoot(printString("\" ++ [",
			printBadRedex(x,
			 lprint(nameLPrint,xs,ss))));
    }
    else if (whnfHead==nameNil && whnfArgs==0)
	updapRoot(consDQuote,ss);
    else
	updateRoot(printString("\" ++ ",printBadRedex(e,ss)));
}

primFun(primNSPrint) {			/* print string without eval	   */
    Cell e  = primArg(2);
    Cell ss = primArg(1);
    unwind(e);
    if (whnfHead==nameCons && whnfArgs==2) {
	Cell x  = pushed(0);
	Cell xs = pushed(1);
	unwind(x);
	if (isChar(whnfHead) && whnfArgs==0)
	    updateRoot(printSChar(charOf(whnfHead),
		        lprint(nameNSPrint,xs,ss)));
	else
	    updateRoot(printString("\" ++ [",
			print(nameNPrint,MIN_PREC,x,
			 lprint(nameNLPrint,xs,ss))));
    }
    else if (whnfHead==nameNil && whnfArgs==0)
	updapRoot(consDQuote,ss);
    else
	updateRoot(printString("\" ++ ",print(nameNPrint,FUN_PREC-1,e,ss)));
}

/* --------------------------------------------------------------------------
 * Auxiliary functions for printer(s):
 * ------------------------------------------------------------------------*/

static Cell local textAsVar(t,ss)	/* reverse t as function symbol	   */
Text t;					/* onto output ss		   */
Cell ss; {
    String s = textToStr(t);
    if ((isascii(s[0]) && isalpha(s[0])) || s[0]=='_' || strcmp(s,"[]")==0)
	return stringOutput(s,ss);
    else
	return ap(consClose,stringOutput(s,ap(consOpen,ss)));
}

static Cell local textAsOp(t,ss)	/* reverse t as op. symbol onto ss */
Text t;
Cell ss; {
    String s = textToStr(t);
    if (isascii(s[0]) && isalpha(s[0]))
	return ap(consBack,stringOutput(s,ap(consBack,ss)));
    else
	return stringOutput(s,ss);
}

static Cell local stringOutput(s,ss)	/* reverse string s onto output ss */
String s;
Cell   ss; {
    while (*s)
	ss = ap(consChar(*s++),ss);
    return ss;
}

static Cell local printBadRedex(rx,rs)	/* Produce expression to print bad */
Cell rx, rs; {				/* redex and then print rest ...   */
    return ap(consObrace,
	    print(nameNPrint,MIN_PREC,rx,
	     ap(consCbrace,
	      rs)));
}

Void abandon(what,rx)			/* abandon computation		   */
String what;
Cell   rx; {
    outputString(errorStream,
		 revOnto(stringOutput("\n",NIL),
  		 revOnto(stringOutput(what,NIL),
 		 revOnto(stringOutput(" error: ",NIL),
			 printDBadRedex(rx,nameNil)))),TRUE);
    errAbort();
}

/* --------------------------------------------------------------------------
 * Evaluate name, obtaining a C string from a Gofer string:
 * ------------------------------------------------------------------------*/

static String local evalName(es)	/* evaluate es :: [Char] and save  */
Cell es; {				/* in char array... return ptr to  */
    static char buffer[FILENAME_MAX+1];	/* string or 0, if error occurs	   */
    Int         pos    = 0;
    StackPtr    saveSp = sp;

    while (isNull(evalWithNoError(es)))
	if (whnfHead==nameCons && whnfArgs==2) {
	    Cell e = pop();		/* avoid leaving anything on stack */
	    es	   = pop();
	    if (isNull(evalWithNoError(e))
			&& isChar(whnfHead) && whnfArgs==0
			&& pos<FILENAME_MAX)
		buffer[pos++] = charOf(whnfHead);
	    else
		break;
	}
	else if (whnfHead==nameNil && whnfArgs==0) {
	    buffer[pos] = '\0';
	    return buffer;
	}
	else
	    break;

    sp = saveSp;			/* stack pointer must be the same  */
    return 0;				/* as it was on entry		   */
}

#if MAC
/* --------------------------------------------------------------------------
 * Call an OS function from Gofer
 * ------------------------------------------------------------------------*/

#if THINKC
pascal long ToolboxTrap  (unsigned short trap,
			  unsigned short ressize,
                          unsigned short args,
                          unsigned short *addr)
{}
#else
extern pascal long ToolboxTrap  (unsigned short trap,unsigned short ressize,unsigned short args,unsigned short *addr);
#endif

extern unsigned long trapreg_d0, trapreg_d1, trapreg_a0, trapreg_a1;

primFun(primTrapReg) {			/* Trap primitive with register settings   */
    Cell trap    = primArg(6);		/* Result always in D0 */
    Cell d0      = primArg(5);
    Cell d1      = primArg(4);
    Cell a0      = primArg(3);
    Cell a1      = primArg(2);
    Cell l       = primArg(1);
    int i;
    
    if(isNull(evalWithNoError(d0)) && isInt(whnfHead) && whnfArgs == 0)
      {
        trapreg_d0 = intOf(whnfHead);
        if(isNull(evalWithNoError(d1)) && isInt(whnfHead) && whnfArgs == 0)
          {
             trapreg_d1 = intOf(whnfHead);
             if(isNull(evalWithNoError(a0)) && isInt(whnfHead) && whnfArgs == 0)
              {
                 trapreg_a0 = intOf(whnfHead);
                 if(isNull(evalWithNoError(a1)) && isInt(whnfHead) && whnfArgs == 0)
                   {
                     trapreg_a1 = intOf(whnfHead);
                     dotrap(root,NIL,trap,l);
		     return;
		   }
	      }
	   }
      }
   updapRoot(ap(ap(ap(ap(ap(nameTrapReg,trap),d0),d1),a0),a1),l);
}


primFun(primTrap) {			/* Trap primitive		   */
    Cell result  = primArg(3);
    Cell trap    = primArg(2);
    Cell l       = primArg(1);
    dotrap(root,result,trap,l);
}
 
 
dotrap(root,result,trap,l)
Cell root,result,trap,l;
{   
    unsigned short buffer[128];
    int cnt    = 0;
    
    if(isNull(evalWithNoError(trap)) && isInt(whnfHead) && whnfArgs == 0)
      {
        int ttrap = (unsigned short) intOf(whnfHead);
        if(isNull(result) || (isNull(evalWithNoError(result)) && isInt(whnfHead) && whnfArgs == 0))
	  {
	    int reskind = isNull(result)?0:intOf(whnfHead);
    
            while (isNull(evalWithNoError(l)))

	      if (whnfHead==nameCons && whnfArgs==2)
	        {
	          Cell e = pop();		/* avoid leaving anything on stack */
	          l	   = pop();
	          if (isNull(evalWithNoError(e))
		      && isInt(whnfHead) && whnfArgs==0
	              && cnt<127)
		      buffer[cnt++] = (unsigned short) intOf(whnfHead);
	          else
		    break;
	        }

	      else if (whnfHead==nameNil && whnfArgs==0)
	        {
		  unsigned short *bp = buffer;
	          updateRoot(mkInt(ToolboxTrap((unsigned short)ttrap,(unsigned short)reskind,cnt,bp)));
	          return;
	        }
	      else
	        break;
          }
      }
    updapRoot(ap(ap(nameTrap,result),trap),l);
}

primFun(primSeq) {			/* Seq primitive		   */
    Cell result  = primArg(1);
    Cell seq     = primArg(2);
    if(isNull(evalWithNoError(seq)))
       updateRoot(result);
    else
      updapRoot(ap(nameSeq,seq),result);
}


primFun(primTrace) {			/* Trace primitive		   */
    Cell result  = primArg(1);
    Cell trace   = primArg(2);
    if(isNull(evalWithNoError(trace)))
      {
         fprintf(stderr,"\nTrace: ");
         printExp(stderr,trace);
         fputc('\n',stderr);
         updateRoot(result);
      }
    else
      updapRoot(ap(nameTrace,trace),result);
}


/*
	unitIO x =        IO ( \ t -> x `seq` (x,t))

	IO u `bindIO` k = IO ( \ t -> let (x,t') =      u t  in
                                      let eval (IO f) = f t' in
                                      x `seq` eval (k x))
*/

primFun(primUnitIO) {
    Cell action = primArg(1);
    Cell result = ap(nameIO,ap(nameCUnitIO,action));
    dotrace("Unitio:   Action =  ",action);
    updateRoot(result);
}

primFun(primCUnitIO) {			/* unitIO primitive -- inner part   */
    Cell token   = primArg(1);
    Cell action  = primArg(2);
    
    dotrace("CUnitio:  Token =   ",token);
    dotrace("CUnitio:  Action =  ",action);

    /* We don't need to evaluate the token: it must be in normal form */
    if(isNull(evalWithNoError(action)))
      updateRoot(buildTuple(cons(token,cons(action,NIL))));
    else
      updapRoot(nameUnitIO,action);
}

/*
unitIO :: a -> IO a
unitIO x = IO ( \ t  -> x `seq` (x, t) )

bindIO :: IO a -> (a -> IO b) -> IO b
IO u `bindIO` k = IO ( \ t -> let (x,t') =      u t  in
                              let eval (IO f) = f t' in
                              x `seq` eval (k x))
*/


primFun(primBindIO) {
    Cell cont =   primArg(1);
    Cell action = primArg(2);

    dotrace("Bindio:   Action =  ",action);
    dotrace("Bindio:   Cont =    ",cont);

    if(isNull(evalWithNoError(action)))
      {
        action = pop();
	cont = ap(nameIO,ap(ap(nameCBindIO,action),cont));
	
        dotrace("Bindio:   Action2 = ",action);
        dotrace("Bindio:   Cont2 =   ",cont);

	/* Eval will force this without growing the C stack */	
	updateRoot(cont);
      }
    else
      updapRoot(ap(nameBindIO,action),cont);
}


primFun(primCBindIO) {			/* bindIO primitive -- inner part   */
    Cell token   = primArg(1);
    Cell cont    = primArg(2);
    Cell action  = primArg(3);
    Cell action2 = ap(action,token);
    
    dotrace("CBindio:  Token =   ",token);
    dotrace("CBindio:  Cont =    ",cont);
    dotrace("CBindio:  Action =  ",action);

    /* Stub out for GC */    
    action = NIL;
    
    if(isNull(evalWithNoError(action2)))
      {
	 action2 = pop();
	 token =  pop();

         dotrace("CBindio:  Token2 =  ",token);
         dotrace("CBindio:  Action2 = ",action2);

	 if(isNull(evalWithNoError(action2)))
	   {
             dotrace("CBindio:  Action3 = ",action2);
	     updapRoot(ap(nameCCBindIO,ap(cont,action2)),token);
	     return;
	   }
      }

    updapRoot(ap(nameBindIO,action),cont);
}


primFun(primCCBindIO) {
    Cell token = primArg(1);
    Cell action = primArg(2);

    dotrace("CCBindio: Action =  ",action);
    dotrace("CCBindio: Token =   ",token);
    
    if(isNull(evalWithNoError(action)))
     {
	action = pop();

        dotrace("CCBindio: Action2 = ",action);
    
	updapRoot(action,token);
      }
    else
      updapRoot(nameBindIO,action);
}   


primFun(primAssign) {			/* Assign primitive		   */
    Cell value  = primArg(1);
    Cell ptr    = primArg(2);
    if(isNull(evalWithNoError(ptr))     && 
       isInt(whnfHead) && whnfArgs == 0)
         {
           Cell eptr = whnfHead;
           if(isNull(evalWithNoError(value))   &&
              isInt(whnfHead) && whnfArgs == 0 )
           {
             (*(int *)intOf(eptr)) = intOf(whnfHead);
             updateRoot(mkInt(0));
           }
	 }
    else
      updapRoot(ap(nameAssign,ptr),value);
}


primFun(primAssignS) {			/* Short Assign primitive	   */
    Cell value  = primArg(1);
    Cell ptr    = primArg(2);
    Cell eptr   = evalWithNoError(ptr);
    Cell evalue; 

    if(isNull(eptr) && isInt(whnfHead) && whnfArgs == 0)
       if(isNull(evalue=evalWithNoError(value))   &&
            isInt(whnfHead) && whnfArgs == 0 )
         {
           (*(short *)intOf(eptr)) = (short) intOf(evalue);
           updateRoot(mkInt(0));
         }
    else
      updapRoot(ap(nameAssignS,ptr),value);
}


primFun(primAssignC) {			/* Char Assign primitive	   */
    Cell value  = primArg(1);
    Cell ptr     = primArg(2);
    if(isNull(evalWithNoError(ptr))     && 
       isInt(whnfHead) && whnfArgs == 0)
         {
           Cell eptr = whnfHead;
           if(isNull(evalWithNoError(value))   &&
              isInt(whnfHead) && whnfArgs == 0 )
           {
             (*(char *)intOf(eptr)) = (char) intOf(whnfHead);
             updateRoot(mkInt(0));
           }
	 }
    else
      updapRoot(ap(nameAssignC,ptr),value);
}


primFun(primAssignBlock) {			/* Short Assign primitive	   */
    Cell values  = primArg(1);
    Cell ptr     = primArg(2);
    
    if(isNull(evalWithNoError(ptr)) && isInt(whnfHead) && whnfArgs == 0)
      {
         short *sptr = (short *) intOf(whnfHead);
    
         while (isNull(evalWithNoError(values)))

	   if (whnfHead==nameCons && whnfArgs==2)
	     {
	       Cell e = pop();		/* avoid leaving anything on stack */
	       values = pop();
	       if (isNull(evalWithNoError(e)) && isInt(whnfHead) && whnfArgs==0)
		 (*sptr++) = (short) intOf(whnfHead);
	       else
		 break;
	     }

	   else if (whnfHead==nameNil && whnfArgs==0)
	     {
	       updateRoot(mkInt(0));
	       return;
	     }
	   else
	     break;
      }
    updapRoot(ap(nameAssignBlock,ptr),values);
}


primFun(primDeref) {			/* Deref primitive		   */
    Cell ptr  = primArg(1);
    Cell val  = 0;
    if(isNull(evalWithNoError(ptr))  &&
       isInt(whnfHead) && whnfArgs == 0 )
	{
          val = *(int *)(intOf(whnfHead));
          updateRoot(mkInt(val));
	}
    else
      updapRoot(nameDeref,ptr);
}
   

primFun(primMalloc) {			/* Malloc primitive		   */
    Cell size  = primArg(1);
    if(isNull(evalWithNoError(size)) &&
       isInt(whnfHead) && whnfArgs == 0 )
      {
        extern char *malloc();
        int thesize = intOf(whnfHead);
	char *ptr = malloc(thesize);
        updateRoot(mkInt((int)ptr));
      }
    else
      updapRoot(nameMalloc,size);
}

primFun(primFree) {			/* Free primitive		   */
    Cell ptr  = primArg(1);
    if(isNull(evalWithNoError(ptr))  &&
       isInt(whnfHead) && whnfArgs == 0 )
         {
	   free((void *)intOf(whnfHead));
           updateRoot(mkInt(0));
	 }
    else
      updapRoot(nameFree,ptr);
}


/* Now some "essential" ToolBox routines */
primFun(primButton) {			/* Button primitive		   */
    Cell arg  = primArg(1);
    if(isNull(evalWithNoError(arg))  && whnfArgs == 0 )
       updateRoot(mkInt((int)Button()));
    else
      updapRoot(nameButton,arg);
}


primFun(primGetMouse) {			/* GetMouse primitive		   */
#if !THINKC
    Cell arg  = primArg(1);
    if(isNull(evalWithNoError(arg))  && whnfArgs == 0 )
       updateRoot(mkInt((int)GetMouse((Point *)intOf(arg))));
    else
      updapRoot(nameGetMouse,arg);
#endif
}

#if MPW
#pragma segment Builtin3
#endif

primFun(primLineTo) {			/* LineTo primitive	   */
    Cell y  = primArg(1);
    Cell x  = primArg(2);
    
    if(isNull(evalWithNoError(x)) && isInt(whnfHead) && whnfArgs == 0)
      {
         int xvalue = intOf(whnfHead);
         if(isNull(evalWithNoError(y)) && isInt(whnfHead) && whnfArgs == 0)
	   {
	      LineTo(xvalue,intOf(whnfHead));
              updateRoot(UNIT);
	      return;
	   }
      }
    updapRoot(ap(nameLineTo,x),y);
}

primFun(primMoveTo) {			/* MoveTo primitive	   */
    Cell y  = primArg(1);
    Cell x  = primArg(2);
    
    if(isNull(evalWithNoError(x)) && isInt(whnfHead) && whnfArgs == 0)
      {
         int xvalue = intOf(whnfHead);
         if(isNull(evalWithNoError(y)) && isInt(whnfHead) && whnfArgs == 0)
	   {
	      MoveTo(xvalue,intOf(whnfHead));
              updateRoot(UNIT);
	      return;
	   }
      }
    updapRoot(ap(nameMoveTo,x),y);
}


primFun(primGetNextEvt) {			/* Event primitive		   */
    Cell mask  = primArg(1);
    if(isNull(evalWithNoError(mask))  && whnfArgs == 0 )
      primgetnextevent(root,whnfHead,FALSE);
    else
      updapRoot(nameGetNextEvt,mask);
}


primFun(primEvtAvail) {			/* Event available primitive		   */
    Cell mask  = primArg(1);
    if(isNull(evalWithNoError(mask))  && whnfArgs == 0 )
      primgetnextevent(root,whnfHead,TRUE);
    else
      updapRoot(nameEvtAvail,mask);
}

/* Get next event primitive separated because of stack problems */
primgetnextevent(root,evtmask,checkonly)
Cell root, evtmask;
Bool checkonly;
{
   extern EventRecord myEvent;
   List result; 

   /* Get next event into myEvent */
   GetNextKbdEvent(intOf(evtmask),checkonly);
   
   /* buildTuple builds in reverse order! */
   result = cons(mkInt((int)(myEvent.modifiers)),
             cons(mkInt(myEvent.where),
               cons(mkInt(myEvent.when),
                cons(mkInt(myEvent.message),
	         cons(mkInt((int)(myEvent.what)),NIL)))));

   updateRoot(buildTuple(result));
}


/* --------------------------------------------------------------------------
 * Callbacks:
 *
 * A fixed buffer of callback functions is maintained.
 *
 * ------------------------------------------------------------------------*/
 
 #if 1
 #define MAX_CALLBACKS 5
 
 struct Callbacks
   {
     Cell callback;
     int (*cfun)();
   } Callbacks[MAX_CALLBACKS];

pascal int callback_0_0()
{
  return(callback_n(0,0,0));
}

pascal int callback_1_0(arg0)
short arg0;
{
  return(callback_n(0,1,(int)arg0));
}

pascal int callback_2_0(arg0)
int arg0;
{
  return(callback_n(0,2,arg0));
}

pascal int callback_0_1()
{
  return(callback_n(1,0,0));
}

pascal int callback_1_1(arg0)
short arg0;
{
  return(callback_n(1,1,(int)arg0));
}

pascal int callback_2_1(arg0)
int arg0;
{
  return(callback_n(1,2,arg0));
}

pascal int callback_0_2()
{
  return(callback_n(2,0,0));
}

pascal int callback_1_2(arg0)
short arg0;
{
  return(callback_n(2,1,(int)arg0));
}

pascal int callback_2_2(arg0)
int arg0;
{
  return(callback_n(2,2,arg0));
}

pascal int callback_0_3()
{
  return(callback_n(3,0,0));
}

pascal int callback_1_3(arg0)
short arg0;
{
  return(callback_n(3,1,(int)arg0));
}

pascal int callback_2_3(arg0)
int arg0;
{
  return(callback_n(3,2,arg0));
}

pascal int callback_0_4()
{
  return(callback_n(4,0,0));
}

pascal int callback_1_4(arg0)
short arg0;
{
  return(callback_n(4,1,(int)arg0));
}

pascal int callback_2_4(arg0)
int arg0;
{
  return(callback_n(4,2,arg0));
}

int callback_n(n,argc,arg)
int n, argc, arg;
{
  Int i;
  Cell callback = Callbacks[n].callback;

  if(argc == 0)
    callback = ap(callback,UNIT);
  else
    callback = ap(ap(callback,mkInt(arg)),UNIT);

  callback = evalWithNoError(callback);

  if(isNull(callback))
    return(intOf(whnfHead));
  else
    return(0);
}

#if 0
int CallbackFns[2][] =
{
  (int) callback_0_0, (int) callback_1_0, (int) callback_2_0,
  (int) callback_0_1, (int) callback_1_1, (int) callback_2_1,
  (int) callback_0_2, (int) callback_1_2, (int) callback_2_2,
  (int) callback_0_3, (int) callback_1_3, (int) callback_2_3,
  (int) callback_0_4, (int) callback_1_4, (int) callback_2_4
};
#else
int CallbackFns[2][];
#endif

pascal int DummyCallBack()
{
  printf("Too many callbacks used\n");
  return(0);
}
   
primFun(primCreateCallback) {			/* Create Callback primitive		   */
  Cell argsize =  primArg(2);
  Cell callback = primArg(1);

  if(isNull(evalWithNoError(argsize)) && whnfArgs == 0)
    {
      Int i;
      Int args = intOf(whnfHead);

      if(args > 2 || args < 0)
        abandon("Callback",ap(ap(nameCreateCallback,argsize),callback));
      else
        {
          for ( i =0; i < MAX_CALLBACKS; ++i)
            if(Callbacks[i].callback == NULL)
              {
		Callbacks[i].callback = callback;
#if 0
		Callbacks[i].cfun = (int (*)())CallbackFns[args][i];
		updateRoot(mkInt((Int)Callbacks[i].cfun));
#endif
		return;
              }

#if 0
	  updateRoot(mkInt((Int)DummyCallback));
#endif
	}
    }
  else
    updapRoot(ap(nameCreateCallback,argsize),callback);
}

primFun(primDisposeCallback)
{
  Cell callback = primArg(1);
  Int i;
  if(isNull(evalWithNoError(callback))  && whnfArgs == 0 )
    for ( i =0; i < MAX_CALLBACKS; ++i)
      if((int) Callbacks[i].cfun == intOf(whnfHead))
        {
          Callbacks[i].callback = NULL;
	  break;
        }
  else
    updapRoot(nameDisposeCallback,callback);
}
#endif
#endif


/* --------------------------------------------------------------------------
 * Dialogue based input/output:
 *
 * N.B. take care when modifying this code - it is rather delicate and even
 * the simplest of changes might create a nasty space leak... you have been
 * warned (please let me know if you think there already is a space leak!).
 * ------------------------------------------------------------------------*/

static Name nameInput;			/* For reading from stdin	   */

static Bool echoChanged;		/* TRUE => echo changed in dialogue*/
static Bool stdinUsed;			/* TRUE => ReadChan stdin has been */
					/*	   seen in dialogue	   */
static FILE *writingFile = 0;		/* points to file open for writing */


Void dialogue(prog)			/* carry out dialogue ...	   */
Cell prog; {				/* :: Dialog=[Response]->[Request] */
    static String ioerr = "Attempt to read response before request complete";
    Cell tooStrict      = mkStr(findText(ioerr));
    Cell resps		= prog = ap(prog,NIL);
    Cell temp;

    echoChanged = FALSE;
    stdinUsed   = FALSE;
    for (;;) {				/* Keep Responding to Requests	   */
	resps = snd(resps) = ap(nameError,tooStrict);
        clearStack();
	if (nonNull(temp=evalWithNoError(prog)))
	    abandonDialogue(temp);
	else if (whnfHead==nameCons && whnfArgs==2) {
	    if (nonNull(temp=evalWithNoError(pushed(0))))
		abandonDialogue(temp);

	    prog = pushed(1+whnfArgs);

	    if (whnfHead==nameReadFile && whnfArgs==1)
		fst(resps) = ap(nameCons,readFile());
	    else if (whnfHead==nameWriteFile && whnfArgs==2)
		fst(resps) = ap(nameCons,writeFile());
	    else if (whnfHead==nameAppendFile && whnfArgs==2)
		fst(resps) = ap(nameCons,appendFile());
	    else if (whnfHead==nameReadChan && whnfArgs==1)
		fst(resps) = ap(nameCons,readChan());
	    else if (whnfHead==nameAppendChan && whnfArgs==2)
		fst(resps) = ap(nameCons,appendChan());
	    else if (whnfHead==nameEcho && whnfArgs==1)
		fst(resps) = ap(nameCons,echo());
	    else if (whnfHead==nameGetArgs && whnfArgs==0)
		fst(resps) = ap(nameCons,getCLArgs());
	    else if (whnfHead==nameGetProgName && whnfArgs==0)
		fst(resps) = ap(nameCons,getProgName());
	    else if (whnfHead==nameGetEnv && whnfArgs==1)
		fst(resps) = ap(nameCons,getEnv());
#if MAC
            else if (whnfHead==nameImperate && whnfArgs==1)
		fst(resps) = ap(nameCons,imperate());
#endif
	    else
		abandonDialogue(pushed(whnfArgs));
	}
	else if (whnfHead==nameNil && whnfArgs==0) {
	    normalTerminal();
	    return;
	}
	else
	    internal("Type error during Dialogue");
    }
}

static Void local abandonDialogue(rx)	/* abandon dialogue after failure  */
Cell rx; {				/* to reduce redex rx		   */
    abandon("Dialogue",rx);
}

static Cell local printDBadRedex(rx,rs) /* Produce expression for bad redex*/
Cell rx, rs; {				/* within a Dialogue, with special */
    if (isAp(rx) && fun(rx)==nameError) /* handling of {error str} redexes */
	return arg(rx);
    else
	return printBadRedex(rx,rs);
}

static Cell local readFile() {		/* repond to ReadFile request	   */
    String s    = evalName(pushed(0));	/* pushed(0) = file name string	   */
    Cell   temp = NIL;			/* pushed(1) = ReadFile request	   */
					/* pushed(2) = rest of program	   */

    if (!s)				/* problem with filename?	   */
	abandonDialogue(pushed(1));
    if (access(s,0)!=0)			/* can't find file		   */ 
	return ap(nameFailure,ap(nameSearchError,pushed(0)));
    if (isNull(temp = openFile(s)))	/* can't open file		   */
	return ap(nameFailure,ap(nameReadError,pushed(0)));
    return ap(nameStr,temp);		/* otherwise we got a file!	   */
}

static Cell local writeFile() {		/* respond to WriteFile req.	   */
    String s    = evalName(pushed(0));	/* pushed(0) = file name string	   */
    FILE   *fp;				/* pushed(1) = output string	   */
    Cell   temp;			/* pushed(2) = output request	   */
					/* pushed(3) = rest of program	   */

    if (!s)				/* problem with filename?          */
        abandonDialogue(pushed(2));
#if MAC
    createTextFile(s);			/* Not automatically created on write */
#endif
    if ((fp=fopen(s,FOPEN_WRITE))==0)	/* problem with output file?	   */
	return ap(nameFailure,ap(nameWriteError,pushed(0)));
    writingFile = fp;
    temp        = outputString(fp,pushed(1),FALSE);
    fclose(fp);
    writingFile = 0;
    if (nonNull(temp))
	return ap(nameFailure,ap(nameWriteError,temp));
    else
	return nameSuccess;
}

static Cell local appendFile() {	/* respond to AppendFile req.	   */
    String s    = evalName(pushed(0));	/* pushed(0) = file name string	   */
    FILE   *fp;				/* pushed(1) = output string	   */
    Cell   temp;			/* pushed(2) = output request	   */
					/* pushed(3) = rest of program	   */

    if (!s)				/* problem with filename?          */
        abandonDialogue(pushed(2));
    if (access(s,0)!=0)			/* can't find file?		   */
	return ap(nameFailure,ap(nameSearchError,pushed(0)));
    if ((fp=fopen(s,FOPEN_APPEND))==0)	/* problem with output file?	   */
	return ap(nameFailure,ap(nameWriteError,pushed(0)));
    writingFile = fp;
    temp        = outputString(fp,pushed(1),FALSE);
    fclose(fp);
    writingFile = 0;
    if (nonNull(temp))
	return ap(nameFailure,ap(nameWriteError,temp));
    else
	return nameSuccess;
}

static Cell local readChan() {		/* respond to readChan req.	   */
    String s    = evalName(pushed(0));	/* pushed(0) = channel name string */
					/* pushed(1) = output request	   */
					/* pushed(2) = rest of program	   */

    if (!s)				/* problem with filename?	   */
	abandonDialogue(pushed(1));
    if (strcmp(s,"stdin")!=0)		/* only valid channel == stdin	   */
	return ap(nameFailure,ap(nameSearchError,pushed(0)));
    if (stdinUsed)			/* can't reuse stdin channel!      */
	return ap(nameFailure,ap(nameReadError,pushed(0)));
    stdinUsed = TRUE;
    return ap(nameStr,ap(nameInput,UNIT));
}

static Cell local appendChan() {	/* respond to AppendChannel req.   */
    String s    = evalName(pushed(0));	/* pushed(0) = channel name string */
    FILE   *fp;				/* pushed(1) = output string	   */
    Cell   temp;			/* pushed(2) = output request	   */
					/* pushed(3) = rest of program	   */

    if (!s)				/* problem with filename?          */
        abandonDialogue(pushed(2));
    if ((fp = validOutChannel(s))==0)	/* problem with output channel?	   */
	return ap(nameFailure,ap(nameSearchError,pushed(0)));
    if (nonNull(temp=outputString(fp,pushed(1),FALSE)))
	return ap(nameFailure,ap(nameWriteError,temp));
    else
	return nameSuccess;
}

static FILE *local validOutChannel(s)	/* return FILE * for valid output  */
String s; {				/* channel name or 0 otherwise...  */
    if (strcmp(s,"stdout")==0)
	return stdout;
    if (strcmp(s,"stderr")==0)
	return stderr;
    if (strcmp(s,"stdecho")==0)		/* in Gofer, stdecho==stdout	   */
	return stdout;
    return 0;
}

#if MAC
extern Bool HandlingEvents;		/* TRUE => Mac I/O taking place	   */

static Cell local imperate() {		/* respond to Imperate request	   */
    					/* pushed(0) = imperative action */
					/* pushed(1) = imperate request	   */
					/* pushed(2) = rest of program	   */
    Cell expr, action = pushed(0);

    /* Set the context for a MacGofer "application" */    
    useprojectresfile(TRUE,FALSE);			/* Use the resources from the Project */
    HandlingEvents = TRUE;
    HideMenus(TRUE);					/* Hide the MacGofer menus	*/
    HideAllWindows(TRUE);				/* And all the windows		*/
    InitCursor();					/* Reset the cursor to an arrow	*/


    /* Evaluate expr in the context of an event-handling program */
    /*  The token version uses UNIT to represent the hidden system state.  */

    expr = evalWithNoError(action);			/* First evaluate the constructor */ 
    /*
       Now do case analysis on the constructor 
       and apply the encapsulated function to the State
    */

     if(isName(whnfHead) && name(whnfHead).defn == CFUN)
      {
        Cell action = pop();				/* Get the action argument from the stack */
        Cell appliedAction = ap(action,UNIT);
        expr = evalWithNoError(appliedAction);
        if(isNull(expr))
	  {
	     Cell state = pop();
	     Cell result = pop();
	     if(isNull(expr=evalWithNoError(state)))
	       expr = evalWithNoError(result);
	  }
      }


    /* Reset the context for a normal Gofer program */
    FlushEvents (everyEvent,0 );                /* Clear all outstanding events		  */
    HideMenus(FALSE);				/* Restore the menus			  */
    HideAllWindows(FALSE);			/* Show all the MacGofer windows	  */
    updatewindows();				/* And redraw them			  */		
    useprojectresfile(FALSE,FALSE);		/* Restore Resources from the application */
    HandlingEvents = FALSE;
    FlushEvents (mDownMask|mUpMask,0 );         /* Clear any mouse events events	  */

    if (isNull(expr))
      return(nameSuccess);
    else
      abandonDialogue(pushed(1));
    return NIL;/*NOTREACHED*/
}

#endif


static Cell local echo() {		/* respond to Echo request	   */
    					/* pushed(0) = boolean echo status */
					/* pushed(1) = echo request	   */
					/* pushed(2) = rest of program	   */
    static String inUse  = "stdin already in use";
    static String repeat = "repeated Echo request";

    if (isNull(evalWithNoError(pushed(0)))) {
	if (stdinUsed)
	    return ap(nameFailure,ap(nameOtherError,mkStr(findText(inUse))));
	if (echoChanged)
	    return ap(nameFailure,ap(nameOtherError,mkStr(findText(repeat))));
	if (whnfHead==nameFalse && whnfArgs==0) {
	    echoChanged = TRUE;
	    noechoTerminal();
	    return nameSuccess;
	}
	if (whnfHead==nameTrue && whnfArgs==0) {
	    echoChanged = TRUE;
	    return nameSuccess;
	}
    }
    abandonDialogue(pushed(1));
    return NIL;/*NOTREACHED*/
}

static Cell local getCLArgs() {		/* get command args -- always []   */
    return ap(nameStrList,nameNil);
}

static Cell local getProgName() {	/* get program name -- an error!   */
    return ap(nameFailure,ap(nameOtherError,nameNil));
}

static Cell local getEnv() {		/* get environment variable	   */
    String s = evalName(pushed(0));	/* pushed(0) = variable name string*/
    String r = 0;			/* pushed(1) = output request	   */
					/* pushed(2) = rest of program	   */
    if (!s)
        abandonDialogue(pushed(1));
    if (r=getenv(s))
	return ap(nameStr,revOnto(stringOutput(r,NIL),nameNil));
    else
	return ap(nameFailure,ap(nameSearchError,pushed(0)));
}

primFun(primInput) {			/* read single character from stdin*/
    Int c = readTerminalChar();

    if (c==EOF || c<0 || c>=NUM_CHARS) {
	clearerr(stdin);
	updateRoot(nameNil);
    }
    else
	updapRoot(consChar(c),ap(nameInput,UNIT));
}

primFun(primFopen) {			/* open file for reading as str	   */
    Cell   succ = primArg(1);		/*  :: String->a->(String->a)->a   */
    Cell   fail = primArg(2);
    String s    = evalName(primArg(3));

    if (s){
	Cell file = openFile(s);
	if (nonNull(file)) {
	    updapRoot(succ,file);
	    return;
	}
    }
    updateRoot(fail);
}

/* --------------------------------------------------------------------------
 * Top-level printing mechanism:
 * ------------------------------------------------------------------------*/

Cell outputString(fp,cs,noDialogue)	/* Evaluate string cs and print	   */
FILE *fp;				/* on specified output stream fp   */
Cell cs;
Bool noDialogue; {			/* TRUE => not runnning Dialogue   */
    Cell temp;

    for (;;) {				/* keep reducing and printing head */
	clearStack();			/* character			   */
	temp = evalWithNoError(cs);
	if (nonNull(temp))
	    if (noDialogue)
		cs = printBadRedex(temp,nameNil);
	    else
		return printDBadRedex(temp,nameNil);
	else if (whnfHead==nameCons && whnfArgs==2) {
	    Cell c = pushed(0);
	    cs     = pushed(1);

	    if (nonNull(temp=evalWithNoError(c)))
		if (noDialogue)
		    cs = printBadRedex(temp,cs);
		else
		    return printDBadRedex(temp,cs);
	    else if (isChar(whnfHead) && whnfArgs==0) {
	        if(!traceEval || !noDialogue) {
		    fputc(charOf(whnfHead),fp);
		    if(!writingFile)
		        fflush(fp);
		}
	    }
	    else
		break;
	}
	else if (whnfHead==nameNil && whnfArgs==0) {
	    if(writingFile)
	       fflush(fp);
	    return NIL;
	}
	else
	    break;
    }
    internal("runtime type error");
    return nameNil;/*NOTREACHED*/
}

/* --------------------------------------------------------------------------
 * Lambda-var prototype implementation:
 * ------------------------------------------------------------------------*/

#ifdef LAMBDAVAR
Void lvExecute(prog)			/* execute lambda var prog of type */
Cell prog; {				/* Proc ()			   */
    Cell temp;
    noechoTerminal();
    temp = evalWithNoError(ap(prog,UNIT));
    if (nonNull(temp))
	abandon("Program execution",temp);
}

primFun(primLvReturn) {			/* lambda var return		   */
    updateRoot(primArg(2));		/* return    :: a -> Proc a	   */
					/* return e _ = e		   */
}

primFun(primLvPure) {			/* lambda var pure		   */
    updapRoot(primArg(1),UNIT);		/* pure  :: Proc a -> a		   */
					/* pure e = e ()		   */
}

primFun(primLvRead) {			/* lambda var reader		   */
    Cell v = primArg(3);		/* (?)::Var a->(a->Proc b)->Proc b */
    Cell f = primArg(2);		/* (Var v ? f) () ===> f v ()	   */
    eval(v);
    if (whnfHead!=nameVar || whnfArgs!=1)
	internal("type error in reader");
    updapRoot(ap(f,pushed(0)),UNIT);
}

primFun(primLvBind) {			/* lambda var bind		   */
    Cell m = primArg(3);		/*($=)::Proc a->(a->Proc b)->Proc b*/
    Cell f = primArg(2);		/* (m $= f) () ===> f (m ()) ()	   */
    Cell a = ap(m,UNIT);		/* strict in first argument	   */
    eval(a);
    updapRoot(ap(f,a),UNIT);
}

primFun(primLvVar) {			/* lambda var, new variable	   */
    updapRoot(ap(primArg(2),		/* var :: (Var a -> Proc b)->Proc b*/
		 ap(nameVar,		/* var f () = f {newvar} ()	   */
		    nameLvUnbound)),
	      UNIT);
}

primFun(primLvNewvar) {			/* lambda var, improved new var	   */
    updapRoot(nameVar,nameLvUnbound);	/* newvar   :: Proc (Var a)	   */
					/* newvar () = {newVar}		   */
}

primFun(primLvAssign) {			/* lambda var assign		   */
    Cell e = primArg(3);		/* assign :: a -> Var a -> Proc () */
    Cell v = primArg(2);	        /* assign e (Var v) () = ()	   */
    eval(v);
    if (whnfHead!=nameVar || whnfArgs!=1)
	internal("type error in assign");
    snd(v) = e;				/* Arrgh! impurity!		   */
    updateRoot(UNIT);
}

primFun(primLvVarEq) {			/* lambda var equality for Vars	   */
    Cell x = primArg(2);		/* :: Var a -> Var a -> Bool	   */
    Cell y = primArg(1);
    eval(x);
    eval(y);				/* I'm not sure this is correct	   */
    updateRoot(x==y ? nameTrue : nameFalse);
}

primFun(primLvGetch) {			/* get character from stdin	   */
    updateRoot(mkChar(readTerminalChar()));
}

primFun(primLvPutchar) {		/* print character on stdout	   */
    eval(primArg(2));			/* putchar c () ==> ()		   */
    putchar(charOf(whnfHead));
    updateRoot(UNIT);
}

primFun(primLvSystem) {			/* do system call		   */
    String s = evalName(primArg(2));	/* system s () ==> int result	   */
    Int    n = s ? system(s) : 1;
    updateRoot(mkInt(n));
}
#endif

/* --------------------------------------------------------------------------
 * Lambda-nu prototype implementation:
 * ------------------------------------------------------------------------*/

#ifdef LAMBDANU
Void lnExecute(prog)			/* execute lambda nu prog of type  */
Cell prog; {				/* Cmd a ()			   */
    Cell temp;
    noechoTerminal();
    temp = evalWithNoError(ap(prog,nameLnDone));
    if (nonNull(temp))
	abandon("Command execution",temp);
}

primFun(primLnDone) {			/* lambda nu done		   */
    updateRoot(UNIT);			/* behaviour is ignored, so isn't  */
}					/* really important		   */

primFun(primLnReturn) {			/* lambda nu return		   */
    updapRoot(primArg(1),primArg(2));	/* return    :: a -> Cmd d a	   */
}					/* return a c = c a		   */

primFun(primLnBind) {			/* lambda nu bind		   */
    Cell a = primArg(3);		/* (>>=)::Cmd c a -> (a -> Cmd c b)*/
    Cell b = primArg(2);		/*			-> Cmd c b */
    Cell c = primArg(1);		/* (a>>=b) c = a (flip b c)	   */
    updapRoot(a,ap(ap(nameLnFlip,b),c));
}

primFun(primLnFlip) {			/* flip primitive, for use in bind */
    updapRoot(ap(primArg(3),primArg(1)),primArg(2));
}

primFun(primLnNew) {			/* lambda nu allocate variable	   */
    Cell c = primArg(1);		/* new :: Cmd a (Tag b)		   */
    updapRoot(c,ap(nameTag,nameLnUnbound));
}

primFun(primLnAssign) {			/* lambda nu assign		   */
    Cell v = primArg(3);		/* assign:: Tag a -> a -> Cmd d () */
    Cell e = primArg(2);	        /* assign (Tag v) e c = c ()	   */
    Cell c = primArg(1);
    eval(v);
    if (whnfHead!=nameTag || whnfArgs!=1)
	internal("type error in assign");
    snd(v) = e;				/* Arrgh! impurity!		   */
    updapRoot(c,UNIT);
}

primFun(primLnRead) {			/* lambda nu reader		   */
    Cell vv = primArg(3);		/* (?) :: Tag a -> (a -> Cmd d b)  */
    Cell b  = primArg(2);		/*			-> Cmd d b */
    Cell c  = primArg(1);		/* (Tag v ? b) c = b v c	   */
    eval(vv);
    if (whnfHead!=nameTag || whnfArgs!=1)
	internal("type error in reader");
    updapRoot(ap(b,pushed(0)),c);
}

primFun(primLnIo) {			/* lambda nu i/o		   */
    updapRoot(primArg(2),primArg(1));	/* io :: ((a->d)->d) -> Cmd d a	   */
}					/* io a c = a c			   */

primFun(primLnBegin) {			/* lambda nu begin		   */
    updapRoot(primArg(1),nameLnNocont);	/* begin :: Cmd d a -> d	   */
}

primFun(primLnTagEq) {			/* lambda nu equality for Tags	   */
    Cell x = primArg(2);		/* :: Tag a -> Tag a -> Bool	   */
    Cell y = primArg(1);
    eval(x);
    eval(y);				/* I'm not sure this is correct	   */
    updateRoot(x==y ? nameTrue : nameFalse);
}

primFun(primLnGetch) {			/* get character from stdin	   */
    updapRoot(primArg(1),mkChar(readTerminalChar()));
}

primFun(primLnPutchar) {		/* print character on stdout	   */
    Cell c = primArg(1);		/* putchar    :: Char -> Cmd a ()  */
    eval(primArg(2));			/* putchar x c = c ()		   */
    putchar(charOf(whnfHead));
    updapRoot(c,UNIT);
}

primFun(primLnSystem) {			/* do system call		   */
    Cell   c = primArg(1);		/* system    :: String -> Cmd a Int*/
    String s = evalName(primArg(2));	/* system s c = c (int result)	   */
    Int    n = s ? system(s) : 1;
    updateRoot(mkInt(n));
}
#endif

#endif

/* --------------------------------------------------------------------------
 * Build array of character conses:
 * ------------------------------------------------------------------------*/

static Cell consCharArray[NUM_CHARS];

Cell consChar(c)			/* return application (:) c	   */
Char c; {
    if (c<0)
	c += NUM_CHARS;
    return consCharArray[c];
}

/*-------------------------------------------------------------------------*/
