# ----------------------------------------------------------------------
# LGRIND makefile for the Macintosh MPW Shell 3.2.
# ----------------------------------------------------------------------
#   Target:     lgrind
#   Sources:    lgrind.c  regexp.c  vgrindefs.c  GUSI.o
#   Created:    Friday, April 30, 1993
# ----------------------------------------------------------------------
# This software was ported to the Macintosh by:
#
# Fred Walsteijn   (E-mail:  walsteyn@fys.ruu.nl)
# Institute for Marine and Atmospheric Research
# Utrecht University
# Princetonplein 5
# 3584 CC Utrecht
# The Netherlands
# ----------------------------------------------------------------------
# The Unix-based code needed only minor changes because the Unix library
# functions from GUSI.o are used.
# GUSI has been made available by Matthias Neeracher (neeri@iis.ethz.ch).
# ----------------------------------------------------------------------
# To recompile this tool, do the following:
#
# 1. Make sure that the variable {MPWversion} has been exported by the Shell.
# 2. Substitute the actual version number of the GUSI library you are using
#    in the following line:
GUSIversion = 1.1.1 
# 3. Change the DEFSFILE definition below to your liking.  A sensible
#    definition would be:   -d DEFSFILE="{MPW}Tools:lgrindefs"
#    Of course you should then put the lgrindefs file in the {MPW}Tools folder.
#    If you're using the def as given below, lgrind expects this file in the
#    current directory if you're not using its -d option.
# 4. After the build of lgrind, use ResEdit to create a new resource in it
#    with type GUI.  The Rez command below has written a template for this
#    type in the resource fork of the tool lgrind.
#    Set the autoSpin value in the new resource to one of 2, 4, 8, 16, or 32;
#    and also set the type and creator of the output file.
#    Change the ID of the new resource to 10240.

VERSION  = 1.0.1
OBJECTS  = lgrind.c.o regexp.c.o vgrindefs.c.o
DEFINES	 = -d MPW  -d DEFSFILE="lgrindefs"
COPTIONS = -model far -mc68020 -mc68881 -elems881 -sym off -mbg off
LOPTIONS = -model far -d -w -c 'MPS ' -t MPST -mf -ad 4 -ac 4 -sym off

lgrind			"{RIncludes}"GUSI.r
	Rez "{RIncludes}"GUSI.r -o lgrind -append    # -d GUSI_PREF_VERSION='0110'

lgrind			{OBJECTS}
	Link {LOPTIONS} 
		 {OBJECTS} 
		"{CLibraries}"GUSI.o 
		"{Libraries}"Stubs.o 
		"{Libraries}"ToolLibs.o 
		"{CLibraries}"Clib881.o 
		"{CLibraries}"CSANELib881.o 
		"{CLibraries}"Math881.o 
		"{CLibraries}"Complex881.o 
		"{CLibraries}"StdClib.o 
		"{Libraries}"Interface.o 
		"{Libraries}"Runtime.o 
		-o lgrind

lgrind.c.o			lgrind.c
	 C {COPTIONS} {DEFINES}  lgrind.c

regexp.c.o			regexp.c
	 C {COPTIONS} {DEFINES}  regexp.c

vgrindefs.c.o		vgrindefs.c
	 C {COPTIONS} {DEFINES}  vgrindefs.c

lgrind				lgrind.make
	SetVersion lgrind -version "{VERSION}" -t vers -i 1 -d 
			-verstring "^,  Fred Walsteijn 1993  (walsteyn@fys.ruu.nl)"
	SetVersion lgrind -version "{MPWversion}" -t vers -i 2 
			-verstring "MPW {MPWversion} / GUSI {GUSIversion}" 
			 Dev:Null || Set Status 0
	# lgrind is built in the current MPW Shell using GUSI of specified version
	# (works only if MPWversion has been exported)
